#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The ADS accessor for the Pandas DataFrame.
The accessor will be initialized with the pandas object the user is interacting with.

Examples
--------
>>> from ads.feature_engineering.accessor.dataframe_accessor import ADSDataFrameAccessor
    >>> from ads.feature_engineering.feature_type.continuous import Continuous
    >>> from ads.feature_engineering.feature_type.creditcard import CreditCard
    >>> from ads.feature_engineering.feature_type.string import String
    >>> from ads.feature_engineering.feature_type.base import Tag
>>> df = pd.DataFrame({'Name': ['Alex'], 'CreditCard': ["4532640527811543"]})
>>> df.ads.feature_type
{'Name': ['string'], 'Credit Card': ['string']}
>>> df.ads.feature_type_description
          Column   Feature Type                        Description
------------------------------------------------------------------
0           Name         string    Type representing string values.
1    Credit Card         string    Type representing string values.
>>> df.ads.default_type
{'Name': 'string', 'Credit Card': 'string'}
>>> df.ads.feature_type = {'Name':['string', Tag('abc')]}
>>> df.ads.tags
{'Name': ['abc']}
>>> df.ads.feature_type = {'Credit Card':['credit_card']}
>>> df.ads.feature_select(include=['credit_card'])
                    Credit Card
-------------------------------
0	          4532640527811543
"""

from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
from ads.common.utils import DATA_SCHEMA_MAX_COL_NUM
from ads.data_labeling.mixin.data_labeling import DataLabelingAccessMixin
from ads.dbmixin.db_pandas_accessor import DBAccessMixin
from ads.feature_engineering import schema
from ads.feature_engineering.accessor.mixin.eda_mixin import EDAMixin
from ads.feature_engineering.accessor.mixin.feature_types_mixin import (
    ADSFeatureTypesMixin,
)
from ads.feature_engineering.feature_type.base import FeatureType
from pandas.core.dtypes.common import is_list_like


@pd.api.extensions.register_dataframe_accessor("ads")
class ADSDataFrameAccessor(
    ADSFeatureTypesMixin, EDAMixin, DBAccessMixin, DataLabelingAccessMixin
):
    """ADS accessor for the Pandas DataFrame.

    Attributes
    ----------
    columns: List[str]
        The column labels of the DataFrame.

    tags(self) -> Dict[str, str]
        Gets the dictionary of user defined tags for the dataframe.
    default_type(self) -> Dict[str, str]
        Gets the map of columns and associated default feature type names.
    feature_type(self) -> Dict[str, List[str]]
        Gets the list of registered feature types.
    feature_type_description(self) -> pd.DataFrame
        Gets the list of registered feature types in a DataFrame format.

    Methods
    -------
    sync(self, src: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame
        Syncs feature types of current DataFrame with that from src.
    feature_select(self, include: List[Union[FeatureType, str]] = None, exclude: List[Union[FeatureType, str]] = None) -> pd.DataFrame
        Gets the list of registered feature types in a DataFrame format.
    help(self, prop: str = None) -> None
        Provids docstring for affordable methods and properties.

    Examples
    --------
    >>> from ads.feature_engineering.accessor.dataframe_accessor import ADSDataFrameAccessor
    >>> from ads.feature_engineering.feature_type.continuous import Continuous
    >>> from ads.feature_engineering.feature_type.creditcard import CreditCard
    >>> from ads.feature_engineering.feature_type.string import String
    >>> from ads.feature_engineering.feature_type.base import Tag
    df = pd.DataFrame({'Name': ['Alex'], 'CreditCard': ["4532640527811543"]})
    >>> df.ads.feature_type
    {'Name': ['string'], 'Credit Card': ['string']}
    >>> df.ads.feature_type_description
              Column   Feature Type                        Description
    -------------------------------------------------------------------
    0           Name         string    Type representing string values.
    1    Credit Card         string    Type representing string values.
    >>> df.ads.default_type
    {'Name': 'string', 'Credit Card': 'string'}
    >>> df.ads.feature_type = {'Name':['string', Tag('abc')]}
    >>> df.ads.tags
    {'Name': ['abc']}
    >>> df.ads.feature_type = {'Credit Card':['credit_card']}
    >>> df.ads.feature_select(include=['credit_card'])
                       Credit Card
    ------------------------------
    0	          4532640527811543
    """

    def __init__(self, pandas_obj) -> None:
        """Initializes ADS Pandas DataFrame Accessor.

        Parameters
        ----------
        pandas_obj : pandas.DataFrame
            Pandas dataframe

        Raises
        ------
        ValueError
            If provided DataFrame has duplicate columns.
        """
        if len(set(pandas_obj.columns)) != len(pandas_obj.columns):
            raise ValueError(
                "Failed to initialize a DataFrame accessor. " "Duplicate column found."
            )
        self._obj = pandas_obj
        super().__init__()
        self.columns = self._obj.columns
        self._info = None

    def info(self) -> Any:
        """Gets information about the dataframe.

        Returns
        -------
        Any
            The information about the dataframe.
        """
        return self._info

    @property
    def _feature_type(self) -> Dict[str, List[FeatureType]]:
        """Gets the map of columns and associated feature types.
        Key is column name and value is list of feature types.
        """
        return {
            self._obj[col].name: self._obj[col].ads._feature_type for col in self._obj
        }

    @property
    def _default_type(self) -> Dict[str, FeatureType]:
        """Gets the map of columns and associated default feature types.
        Key is column name and value is a default feature type.
        """
        return {
            self._obj[col].name: self._obj[col].ads._default_type for col in self._obj
        }

    @property
    def tags(self) -> Dict[str, List[str]]:
        """Gets the dictionary of user defined tags for the dataframe. Key is column name
        and value is list of tag names.

        Returns
        -------
        Dict[str, List[str]]
            The map of columns and associated default tags.
        """
        return {self._obj[col].name: self._obj[col].ads.tags for col in self._obj}

    @property
    def default_type(self) -> Dict[str, str]:
        """Gets the map of columns and associated default feature type names.

        Returns
        -------
        Dict[str, str]
            The dictionary where key is column name and value is the name of default feature
            type.
        """
        return {k: v.name for k, v in self._default_type.items()}

    @property
    def feature_type(self) -> Dict[str, List[str]]:
        """Gets the list of registered feature types.

        Returns
        -------
        Dict[str, List[str]]
            The dictionary where key is column name and value is list of associated feature type
            names.
        """
        return {col.name: col.ads.feature_type for _, col in self._obj.items()}

    @property
    def feature_type_description(self) -> pd.DataFrame:
        """Gets the list of registered feature types in a DataFrame format.

        Returns
        -------
        :class:`pandas.DataFrame`

        Examples
        ________
        >>> df.ads.feature_type_description()
                  Column   Feature Type                         Description
        -------------------------------------------------------------------
        0           City         string    Type representing string values.
        1   Phone Number         string    Type representing string values.
        """
        result_df = pd.DataFrame([], columns=["Column", "Feature Type", "Description"])
        for col in self._obj:
            series_feature_type_df = self._obj[col].ads.feature_type_description
            series_feature_type_df.insert(0, "Column", col)
            result_df = result_df.append(series_feature_type_df)
        result_df.reset_index(drop=True, inplace=True)
        return result_df

    @feature_type.setter
    def feature_type(
        self, feature_type_map: Dict[str, List[Union[FeatureType, str]]]
    ) -> None:
        """Sets feature types for the DataFrame.

        Parameters
        ----------
        feature_type_map : Dict[str, List[Union[FeatureType, str]]]
            The map of feature types where key is column name and value is list of feature
            types.

        Returns
        -------
        None
            Nothing.
        """
        for col, feature_types in feature_type_map.items():
            self._obj[col].ads.feature_type = feature_types

    def sync(self, src: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
        """Syncs feature types of current DataFrame with that from src.

        Syncs feature types of current dataframe with that from src, where src
        can be a dataframe or a series. In either case, only columns with
        matched names are synced.

        Parameters
        ----------
        src: `pd.DataFrame` | `pd.Series`
            The source to sync from.

        Returns
        -------
        :class:`pandas.DataFrame`
            Synced dataframe.
        """
        for _, col in self._obj.items():
            col.ads.sync(src)

    def _extract_columns_of_target_types(
        self, target_types: List[Union[FeatureType, str]]
    ) -> List:
        """Returns all the column names that are of the target types from the
        feature_type dictionary.

        Parameters
        ----------
        target_types: list
            A list of target feature types, can be either feature type names of
            feature type class.

        Returns:
        -------
        List[str]
            The list of columns names.
        """
        columns = []
        target_types = (
            np.unique(
                [self._get_type(feature_type).name for feature_type in target_types]
            )
            if target_types is not None
            else None
        )
        for target_type in target_types:
            for name, feature_types in self.feature_type.items():
                if target_type in feature_types:
                    columns.append(name)
        return columns

    def feature_select(
        self,
        include: List[Union[FeatureType, str]] = None,
        exclude: List[Union[FeatureType, str]] = None,
    ) -> pd.DataFrame:
        """Returns a subset of the DataFrame’s columns based on the column feature_types.

        Parameters
        ----------
        include: List[Union[FeatureType, str]], optional
            Defaults to None. A list of FeatureType subclass or str to be included.
        exclude: List[Union[FeatureType, str]], optional
            Defaults to None. A list of FeatureType subclass or str to be excluded.

        Raises
        ------
        ValueError
            If both of include and exclude are empty
        ValueError
            If include and exclude are used simultaneously

        Returns
        -------
        :class:`pandas.DataFrame`
            The subset of the frame including the feature types in include and excluding
            the feature types in exclude.
        """
        if not (include or exclude):
            raise ValueError("at least one of include or exclude must be nonempty")

        if not is_list_like(include):
            include = (include,) if include is not None else ()
        if not is_list_like(exclude):
            exclude = (exclude,) if exclude is not None else ()

        # unify the feature types to str representation
        include = (
            np.unique([self._get_type(feature_type).name for feature_type in include])
            if include is not None
            else None
        )
        exclude = (
            np.unique([self._get_type(feature_type).name for feature_type in exclude])
            if exclude is not None
            else None
        )

        # convert the myriad valid dtypes object to a single representation
        include = frozenset(include)
        exclude = frozenset(exclude)

        # can't both include AND exclude!
        if not include.isdisjoint(exclude):
            raise ValueError(f"include and exclude overlap on {(include & exclude)}")

        # We raise when both include and exclude are empty
        # Hence, we can just shrink the columns we want to keep
        keep_these = np.full(self._obj.shape[1], True)

        columns = self._obj.columns

        if include:
            included_columns = self._extract_columns_of_target_types(include)
            keep_these &= columns.isin(included_columns)

        if exclude:
            excluded_columns = self._extract_columns_of_target_types(exclude)
            keep_these &= ~columns.isin(excluded_columns)

        return self._obj.loc[:, keep_these]

    def _add_feature_type(
        self, col: str, feature_type: Union[FeatureType, str]
    ) -> bool:
        """Adds a feature type

        Parameters
        ----------
        col : str
            The column name.
        feature_type : Union[FeatureType, str]
            The feature type to add.

        Returns
        -------
        bool
            Whether add succeeded.
        """
        if col not in self._obj.columns:
            raise ValueError(f"Column {col} is not found.")
        return self._obj[col].ads._add_feature_type(feature_type)

    def _remove_feature_type(
        self, col: str, feature_type: Union[FeatureType, str]
    ) -> None:
        """Removes a feature type

        Parameters
        ----------
        col : str
            column name
        feature_type : Union[FeatureType, str]
            feature type

        Returns
        -------
        None
            Nothing
        """
        if col not in self._obj.columns:
            raise ValueError(f"Column {col} is not found.")
        self._obj[col].ads._remove_feature_type(feature_type)

    def model_schema(self, max_col_num: int = DATA_SCHEMA_MAX_COL_NUM):
        """
        Generates schema from the dataframe.

        Parameters
        ----------
        max_col_num : int, optional. Defaults to 1000
            The maximum column size of the data that allows to auto generate schema.

        Examples
        --------
        >>> df = pd.read_csv('./orcl_attrition.csv', usecols=['Age', 'Attrition'])
        >>> schema = df.ads.model_schema()
        >>> schema
        Schema:
            - description: Attrition
            domain:
                constraints: []
                stats:
                count: 1470
                unique: 2
                values: String
            dtype: object
            feature_type: String
            name: Attrition
            required: true
            - description: Age
            domain:
                constraints: []
                stats:
                25%: 31.0
                50%: 37.0
                75%: 44.0
                count: 1470.0
                max: 61.0
                mean: 37.923809523809524
                min: 19.0
                std: 9.135373489136732
                values: Integer
            dtype: int64
            feature_type: Integer
            name: Age
            required: true
        >>> schema.to_dict()
        {'Schema': [{'dtype': 'object',
            'feature_type': 'String',
            'name': 'Attrition',
            'domain': {'values': 'String',
                'stats': {'count': 1470, 'unique': 2},
                'constraints': []},
            'required': True,
            'description': 'Attrition'},
            {'dtype': 'int64',
            'feature_type': 'Integer',
            'name': 'Age',
            'domain': {'values': 'Integer',
                'stats': {'count': 1470.0,
                'mean': 37.923809523809524,
                'std': 9.135373489136732,
                'min': 19.0,
                '25%': 31.0,
                '50%': 37.0,
                '75%': 44.0,
                'max': 61.0},
                'constraints': []},
            'required': True,
            'description': 'Age'}]}

        Returns
        -------
        ads.feature_engineering.schema.Schema
            data schema.

        Raises
        ------
        ads.feature_engineering.schema.DataSizeTooWide
            If the number of columns of input data exceeds `max_col_num`.
        """
        if max_col_num and len(self._obj.columns) > max_col_num:
            raise schema.DataSizeTooWide(
                data_col_num=len(self._obj.columns), max_col_num=max_col_num
            )

        sc = schema.Schema()
        for i, col in enumerate(self._obj.columns):
            domain = schema.Domain()
            try:
                domain = self._obj[col].ads.feature_domain()
            except:
                pass

            sc.add(
                schema.Attribute(
                    self._obj[col].dtype.name,
                    domain.values,
                    col,
                    domain=domain,
                    description=str(col),
                    required=bool(~self._obj[col].isnull().any()),
                    order=i,
                )
            )

        return sc

    def __getattr__(self, attr):
        attr_map = dict()
        for col in self._obj:
            try:
                val = self._obj[col].ads.__getattr__(attr)
            except:
                val = None  # if a column does not have the request attr, return None
            attr_map[col] = val

        if any(
            callable(x) for x in list(attr_map.values())
        ):  # check if attr is a callable, and if yes apply args to all cols.

            def func(*args, **kwargs):
                out = dict()
                for k, v in attr_map.items():
                    out[k] = v(*args, **kwargs) if v else None
                return out

            return func

        return attr_map



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ads
from ads.common.utils import MYSQL_DEFAULT_PORT
from ads.secrets import SecretKeeper, Secret
import json


logger = ads.getLogger("ads.secrets")

from dataclasses import dataclass, field


@dataclass
class MySQLDBSecret(Secret):
    """
    Dataclass representing the attributes managed and serialized by MySQLDBSecretKeeper
    """

    user_name: str
    password: str
    host: str
    port: str
    database: str = field(default=None)


class MySQLDBSecretKeeper(SecretKeeper):
    """
    `MySQLDBSecretKeeper` provides an interface to save MySQL database credentials.
    If you use Wallet file for connnecting to the database, please use ``ADBSecretKeeper``.


    Examples
    --------
    >>> from ads.secrets.mysqldb import MySQLDBSecretKeeper
    >>> vault_id = "ocid1.vault.oc1..<unique_ID>"
    >>> key_id = "ocid1.key..<unique_ID>"

    >>> import ads
    >>> ads.set_auth("resource_principal") # If using resource principal for authentication
    >>> connection_parameters={
    ...     "user_name":"<your user name>",
    ...     "password":"<your password>",
    ...     "host":"<db host>",
    ...     "port":"<db port>",
    ...     "database":"<database>",
    ... }
    >>> mysqldb_keeper = MySQLDBSecretKeeper(vault_id=vault_id, key_id=key_id, **connection_parameters)
    >>> mysqldb_keeper.save("mysqldb_employee", "My DB credentials", freeform_tags={"schema":"emp"})
    >>> print(mysqldb_keeper.secret_id) # Prints the secret_id of the stored credentials
    >>> mysqldb_keeper.export_vault_details("mysqldb_employee_att.json") # Save the secret id and vault info to a json file

    >>> # Loading credentails
    >>> import ads
    >>> ads.set_auth("resource_principal") # If using resource principal for authentication
    >>> from ads.secrets.mysqldb import MySQLDBSecretKeeper
    >>> secret_id = "ocid1.vaultsecret.oc1..<unique_ID>"
    >>> with MySQLDBSecretKeeper.load_secret(source=secret_id) as mysqldb_creds:
    ...     import pandas as pd
    ...     df = pd.DataFrame.ads.read_sql("select * from EMPLOYEE", connection_parameters=mysqldb_creds, engine="mysql")


    >>> mymysqldb_creds = MySQLDBSecretKeeper.load_secret(source='mysqldb_employee_att.json', format="json")
    >>> pd.DataFrame.ads.read_sql("select * from ATTRITION_DATA", connection_parameters=mymysqldb_creds.to_dict(), engine="mysql").head(2)

    """

    def __init__(
        self,
        user_name: str = None,
        password: str = None,
        host: str = None,
        port: str = MYSQL_DEFAULT_PORT,
        database: str = None,
        repository_path: str = None,
        repository_key: str = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        user_name: (str, optional). Default None
            user_name of the database
        password: (str, optional). Default None
            password  for connecting to the database
        host: (str, optional). Default None
            Database host name
        port: (str, optional). Default 1521
            Port number
        database: (str, optional). Default None
            database name
        repository_path: (str, optional). Default None.
            Path to credentials repository. For more details refer `ads.database.connection`
        repository_key: (str, optional). Default None.
            Configuration key for loading the right configuration from repository. For more details refer `ads.database.connection`
        kwargs:
            vault_id: str. OCID of the vault where the secret is stored. Required for saving secret.
            key_id: str. OCID of the key used for encrypting the secret. Required for saving secret.
            compartment_id: str. OCID of the compartment where the vault is located. Required for saving secret.
            auth: dict. Dictionay returned from ads.common.auth.api_keys() or ads.common.auth.resource_principal(). By default, will follow what is set in `ads.set_auth`.  Use this attribute to override the default.
        """
        if repository_path and repository_key:
            from ads.database import connection

            config_from_repo = connection.get_repository(
                repository_key, repository_path
            )
            if user_name is None:
                user_name = config_from_repo.get("username") or config_from_repo.get(
                    "user_name"
                )
            if password is None:
                password = config_from_repo.get("password")
            if database is None:
                database = config_from_repo.get("database")
            if host is None:
                host = config_from_repo.get("host")
            if port is None:
                port = config_from_repo.get("port")

        self.data = MySQLDBSecret(
            user_name=user_name,
            password=password,
            host=host,
            port=port,
            database=database,
        )

        super().__init__(**kwargs)

    def decode(self) -> "ads.secrets.mysqldb.MySQLDBSecretKeeper":
        """
        Converts the content in `self.encoded` to `MySQLDBSecret` and stores in `self.data`

        Returns
        -------
        MySQLDBSecretKeeper:
            Returns self object
        """
        content = json.loads(self._decode())
        logger.debug(f"Decoded secret contains following keys - {list(content.keys())}")
        self.data = MySQLDBSecret(
            **content,
        )
        return self



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ipaddress
import json
import os
import sys
from time import sleep, time, time_ns
from urllib.parse import urlparse
import subprocess


import ads
import fsspec
import oci
import pandas as pd  # Have to find a better way for timedelta
import psutil
from ads.jobs import Job
from ads.opctl.distributed.common import cluster_config_helper
from ads.jobs.templates.driver_oci import GitSSHKey, GitManager, JobRunner
from ads.jobs.builders.runtimes.artifact import Artifact


class ClusterProvider:
    """
    Provides contract for implementing Framework specific Cluster Life Cycle Manager.

    The main node is identified by the environment variable - `MAIN`
    The worker node is identified by the environment variable - `WORKER`

    The worker and main coordinate cluster configuration using the directory provided via `WORK_DIR`. The main node creates config.json in the `WORK_DIR`.
    The worker nodes polls the config.json and exports the configuration as environment variables
    """

    SYNC_SCRIPT_PATH = "/etc/datascience/sync.sh"

    DEFAULT_CODE_DIR = "/code"

    def __init__(self, mode, ephemeral=True, life_span="0h", work_dir=""):
        self.mode = mode
        self.start_time = time()
        self.ephemeral = ephemeral
        self.authinfo = self.get_oci_auth()
        self.ip = self.find_self_ip(self.authinfo)
        self.end_time = (
            pd.Timedelta(int(life_span[:-1]), life_span[-1]).total_seconds()
            + self.start_time
        )
        self.work_dir = work_dir
        self._get_my_work_dir()

        self.main_config_file = os.path.join(
            self.my_work_dir, f"MAIN_config.json"
        )  # In case of worker, we have two config file, worker generated and main config
        self.time_out = int(
            os.environ.get("OCI__TIMEOUT", "600")
        )  # Time out to wait for the config file

        self.fetch_code()

        self.setup_configuration()  # Write cluster configuration to `work_dir`. Eg. IP address of scheduler, etc.

        # self.tmpdir = os.environ["tmpdir"]

        self.stop_file = os.path.join(
            self.my_work_dir, "stop"
        )  # Control file to instruct cluster to stop

        scheme = urlparse(
            self.stop_file,
        ).scheme
        self.scheme = scheme

        self.execution_failure = False
        self.code_execution_complete = False
        self.sync()

    def sync(self, loop=True):
        sync_artifacts = os.environ.get("SYNC_ARTIFACTS", 0)
        print(f"sync_artifacts - {sync_artifacts}")
        if sync_artifacts == "1":
            bkt_name, prefix = self.get_sync_loc()
            if bkt_name is None:
                print(
                    "WARNING: Sync 'WORKSPACE', 'WORKSPACE_PREFIX' or 'work_dir' not configured. Skipping Sync"
                )
                return
            sync_script_fn = self.SYNC_SCRIPT_PATH
            if not os.path.exists(sync_script_fn):
                sync_script = self.get_sync_script()
                self.create_sync_script(sync_script_fn, sync_script)
            subprocess.Popen(
                [sync_script_fn, bkt_name, prefix, "-l"]
            ) if loop else subprocess.Popen([sync_script_fn, bkt_name, prefix])
        if not loop:
            sleep_duration = int(os.environ.get("POST_PROCESSING_WAIT", 60))
            print(f"post processing wait..{sleep_duration} seconds")
            sleep(sleep_duration)
            print("..")

    def get_sync_loc(self):
        bckt_name = os.environ.get("WORKSPACE")
        pfx = os.environ.get("WORKSPACE_PREFIX")
        if bckt_name is None:
            scheme = urlparse(self.work_dir)
            if scheme.scheme == "oci":
                bckt_name = scheme.netloc.split("@")[0]
                pfx = scheme.path
                pfx = pfx.strip("//")
        return bckt_name, pfx

    def profile_cmd(self):
        profile = os.environ.get("PROFILE", "0")
        cmd = []
        if profile == "1":
            print("Profiler ON")
            cmd = os.environ.get("PROFILE_CMD").split(" ")
        return cmd

    @staticmethod
    def create_sync_script(sync_script_fn, sync_script):
        sync_script_fn_obj = open(sync_script_fn, "w")
        sync_script_fn_obj.write(sync_script)
        sync_script_fn_obj.close()
        os.chmod(sync_script_fn, 755)

    def get_sync_script(self):
        return sync_script_str

    def fetch_code(self):
        runtime_type = os.environ.get(cluster_config_helper.OCI__RUNTIME_TYPE)
        if not runtime_type:
            return
        delegates = {"git": self._fetch_git, "remote": self._fetch_remote}
        if runtime_type not in delegates:
            raise ValueError(f"Runtime type {runtime_type} is not supported.")
        if cluster_config_helper.OCI__CODE_DIR not in os.environ:
            os.environ[cluster_config_helper.OCI__CODE_DIR] = self.DEFAULT_CODE_DIR
        delegates[runtime_type](
            code_dir=os.environ.get(cluster_config_helper.OCI__CODE_DIR)
        )

    def _fetch_git(self, code_dir):
        uri = os.environ.get(cluster_config_helper.OCI__RUNTIME_URI)
        branch = os.environ.get(cluster_config_helper.OCI__RUNTIME_GIT_BRANCH)
        commit = os.environ.get(cluster_config_helper.OCI__RUNTIME_GIT_COMMIT)
        secret_ocid = os.environ.get(cluster_config_helper.OCI__RUNTIME_GIT_SECRET_ID)
        # with GitSSHKey does nothing if secret_ocid is None or empty
        with GitSSHKey(secret_ocid):
            GitManager(uri, code_dir=code_dir).fetch_repo().checkout_code(
                branch=branch, commit=commit
            )
        JobRunner(code_dir=code_dir).setup_python_path(
            python_path=os.environ.get(cluster_config_helper.OCI__RUNTIME_PYTHON_PATH),
        )

    def _fetch_remote(self, code_dir):
        uri = os.environ.get(cluster_config_helper.OCI__RUNTIME_URI)
        Artifact.copy_from_uri(uri, code_dir)

    def run_code(self):
        # Sub-class should implement this method to run code and,
        # set code_execution_complete to True after running code.
        self.code_execution_complete = True

    @property
    def stop_filesystem(self):
        authinfo = {}
        if self.scheme == "oci":
            authinfo = self.get_oci_auth()
        return fsspec.filesystem(
            self.scheme, **authinfo
        )  # FileSystem class corresponding to the URI scheme.

    def _get_my_work_dir(self):
        """
        Get the work dir subfolder for the current running job
        """
        # Use "Undefined" as job identifier when not running on OCI jobs.
        self.jobDefID = os.environ.get("JOB_OCID", "Undefined")
        # Use time as job run identifier when not running on OCI jobs
        self.jobRunID = os.environ.get("JOB_RUN_OCID", str(time_ns()))

        self.my_work_dir = os.path.join(self.work_dir, self.jobDefID)
        self.config_file = os.path.join(
            self.my_work_dir,
            "MAIN_config.json"
            if self.mode == "MAIN"
            else f"{self.mode}_{self.jobRunID}_config.json",
        )  # Config file location

    def export_config_files(self):
        """
        By default only exports configuration generated by main. This behavior can be overridden.
        """
        return [self.main_config_file if self.mode == "WORKER" else self.config_file]

    def reached_endoflife(self):
        # TODO We can get rid of this method as JobRun takes parameter to set
        # the max run time. We dont have to handle this here.
        return (self.end_time - self.start_time) <= 0

    def get_oci_auth(self):
        profile = os.environ.get("OCI_CONFIG_PROFILE") or os.environ.get(
            "OCIFS_CONFIG_PROFILE"
        )
        ads.set_auth(
            os.environ.get("OCI_IAM_TYPE", "resource_principal"),
            profile=profile or "DEFAULT",
        )
        authinfo = ads.common.auth.default_signer()
        return authinfo

    @classmethod
    def find_self_ip(cls, authinfo):
        """
        Identify IP address by finding which of the host IP intersects with the CIDR block of the subnet associated with the JOB_OCID
        """
        if os.environ.get("JOB_OCID"):
            job_ocid = os.environ["JOB_OCID"]
            jobDef = Job.from_datascience_job(job_ocid)
            subnet_id = jobDef.infrastructure.subnet_id
            core_client = oci.core.VirtualNetworkClient(**authinfo)
            cidr = core_client.get_subnet(subnet_id).data.cidr_block

            for interface, snics in psutil.net_if_addrs().items():
                ip = snics[0].address
                if ipaddress.ip_address(ip) in ipaddress.ip_network(cidr):
                    print(f"IP address: {ip}")
                    os.environ["GLOO_SOCKET_IFNAME"] = interface
                    os.environ["NCCL_SOCKET_IFNAME"] = interface
                    return ip
            print("IP ADDRESS NOT FOUND!!")
            return None
        else:
            import socket

            hostname = socket.gethostname()
            ip = socket.gethostbyname(hostname)
            print(f"IP address: {ip}")
            return ip

    def basic_configuration(
        self,
    ) -> dict:
        """
        Prepares basic set of configuration which is framework independent.
        This configuration is decorated later by `configuration` method implemented by framework specific implementations
        """
        config = {}
        config["OCI__MAIN_IP" if self.mode == "MAIN" else "OCI__WORKER_IP"] = self.ip
        config["tmpdir"] = self.my_work_dir
        return config

    def configuration(self, conf={}) -> dict:
        """
        Provides the configuration information of the cluster.

        conf:
            Contains generic information about the cluster, generated using `basic_configuration`. Eg. IP Address of the main process
        """
        return None

    def setup_extra_configs(self, conf: dict):
        return None

    def setup_configuration(self, config: dict = None):
        """
        Writes the configuration information into location provided by `work_dir`

        config:
            dictionary containing configuration information that needs to be shared with the workers
            if config is None, then this method calls `self.configuration` and saves the configuration
        work_dir:
            Could be any valid URI supported by fsspec

        """
        # tmpdir = self.my_work_dir
        # self.tmpdir = tmpdir
        config = config or self.configuration()
        print(
            f"Writing configuration: {config} to file: {self.config_file}", flush=True
        )
        with fsspec.open(self.config_file, "w", **self.authinfo) as conf:
            conf.write(json.dumps(config))

    def export_configuration(self, files):
        """
        Read the configuration in the files array  and export to environment variable
        """
        print(f"{self.mode}. Config File: {files}", flush=True)
        for file in files:
            with fsspec.open(file, **self.authinfo) as conf:
                config = json.loads(conf.read())
                print(f"Loading config: {config}", flush=True)
            for k in config:
                os.environ[k] = str(config[k])

    def expected_worker_count(
        self,
    ):
        return int(os.environ.get("OCI__WORKER_COUNT", 1))

    def fetch_all_worker_info(self):
        """
        Fetchs all the worker configs
        In some cluster the main requires information about all worker IPs apriori. This method maybe useful in such situation.
        """
        files = self.stop_filesystem.ls(
            f"{self.my_work_dir}/WORKER*config.json", refresh=True
        )
        worker_details = {}
        for file in files:
            with open(file) as wcf:
                worker_details[file] = json.loads(wcf.read())
        return worker_details

    def start_main(self):
        """
        Implement this for starting the main process. Eg. `scheduler` for Dask
        """
        pass

    def start_worker(self):
        """
        Implement this for starting the worker process. Eg. `dask-worker` for Dask
        """
        pass

    def start_ps(self):
        """
        Implement this for starting the ps process. Eg. `tf-parameter-server` for tensorflow
        """
        pass

    def start(self):
        """
        Starts the cluster processes
        """
        if self.mode == "MAIN":  # Check if the docker is started in scheduler mode
            print(f"Starting main process", flush=True)
            self.start_main()
        elif self.mode == "WORKER":  # Check if the docker is staretd in worker mode
            print(f"Starting worker process", flush=True)
            self.when_ready(self.start_worker)
        elif self.mode == "PS":  # Check if the docker is staretd in worker mode
            print(f"Starting ps process", flush=True)
            self.when_ready(self.start_ps)
        else:
            print(f"Not a valid mode: {self.mode}", flush=True)
            raise Exception("Not a valid mode")

    def check_cluster_status(self):
        pass

    def execution_failed(self):
        """
        Invoked when code submitted to epheramal cluster fails. Calling this method sets the cluster tearable state
        """
        self.execution_failure = True

    def tearable(self):
        """
        Checks if cluster is ready for tear down.
        If there is a `stop` file in the tmp directory then stop.
        If cluster is ephemeral type, check if code execution is complete
        If TTL is reached then stop
        """

        if self.stop_filesystem.exists(self.stop_file) or self.execution_failure:
            print(
                f"Stopping the process. Reason: {'Code Execution Failure' if self.execution_failure else 'Stop file found'}",
                flush=True,
            )
            return True

        if self.ephemeral:
            return self.code_execution_complete
        else:
            return self.reached_endoflife()

    def stop(self):
        """
        Writes stop file and exits.
        """
        if not self.stop_filesystem.exists(self.stop_file):
            print("Stop file not found. Writing stop file....", flush=True)
            authinfo = {}
            if self.scheme == "oci":
                authinfo = self.get_oci_auth()
            with fsspec.open(self.stop_file, "w", **authinfo) as sf:
                sf.write("stop")

    def ready(self):
        if self.stop_filesystem.exists(self.main_config_file):
            return True
        else:
            return False

    def when_ready(self, func, *args, **kwargs):
        start_time = time()
        while not self.ready():
            sleep(10)
            if time() - start_time > self.time_out:
                raise Exception(
                    "Timed out waiting to be in ready status. Likely cause configuration is missing in the `WORK_DIR`"
                )

        func(*args, **kwargs)


sync_script_str = """#!/bin/bash

loop=false

if [ "$3" == "-l" ]; then
  loop=true;
fi
echo "loop: $loop"
sleep_duration=${SYNC_SLEEP:-60}
if [ "$OCI_IAM_TYPE" = 'api_key' ]; then
  auth_method=api_key
else
  auth_method=resource_principal
fi
echo "auth method: $auth_method"
echo "OCI__SYNC_DIR dir: $OCI__SYNC_DIR"
echo "sleep duration is $sleep_duration"

bucket=$1
prefix=$2/sync/$JOB_OCID/$JOB_RUN_OCID/

while true; do
  if [[ -d $OCI__SYNC_DIR && -n "$(ls -A $OCI__SYNC_DIR)" ]]; then
    echo "syncing $OCI__SYNC_DIR to $bucket/$prefix"
    $HOME/bin/oci os object sync --auth $auth_method --bucket-name $bucket --prefix $prefix --src-dir $OCI__SYNC_DIR
  else
    echo "nothing to sync in $OCI__SYNC_DIR"
  fi
  if [ "$loop" = false ] ; then
    break
  fi
  sleep $sleep_duration
done
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from email.policy import default
import logging
from unittest.mock import DEFAULT
import numpy as np
import pandas as pd
import re

from abc import ABC, abstractmethod
from time import time
from typing import Dict, Iterator, List, Optional, Union
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)

logger = logging.getLogger("ads.hive_connector")
CURSOR_SIZE = 50000
SERVICE_NAME = "hive"
DEFAULT_BATCH_SIZE = 1000
BDS_HIVE_DEFAULT_PORT = "10000"


class HiveConnection(ABC):
    """Base class Interface."""

    def __init__(self, **params):
        """set up hive connection."""
        self.params = params
        self.con = None  # setup the connection

    @abstractmethod
    def get_cursor(self):
        """Returns the cursor from the connection.

        Returns
        -------
        HiveServer2Cursor:
            cursor using a specific client.
        """
        pass

    @abstractmethod
    def get_engine(self):
        """Returns engine from the connection.

        Returns
        -------
        Engine object for the connection.
        """
        pass


class ImpylaHiveConnection(HiveConnection):
    """ImpalaHiveConnection class which uses impyla client."""

    @runtime_dependency(module="impala", install_from=OptionalDependency.BDS)
    def __init__(self, **params):
        """set up the impala connection."""
        from impala.dbapi import connect

        self.params = params
        self.con = connect(**self.params)

    def get_cursor(self) -> "impala.hiveserver2.HiveServer2Cursor":
        """Returns the cursor from the connection.

        Returns
        -------
        impala.hiveserver2.HiveServer2Cursor:
            cursor using impyla client.
        """
        return self.con.cursor()

    @runtime_dependency(module="sqlalchemy", install_from=OptionalDependency.BDS)
    def get_engine(self, schema="default"):
        """return the sqlalchemy engine from the connection.

        Parameters
        ----------
        schema: str
            Default to "default". The default schema used for query.

        Returns
        -------
        sqlalchemy.engine:
            engine using a specific client.
        """
        from sqlalchemy.engine import create_engine

        logger.info(
            f'Creating sqlalchemy engine with: hive://{self.params["host"]}:{self.params["port"]}/{schema}'
        )
        return create_engine(
            f'hive://{self.params["host"]}:{self.params["port"]}/{schema}'
        )


class HiveConnectionFactory:
    clientprovider = {
        "impyla": ImpylaHiveConnection,
    }

    @classmethod
    def get(cls, driver="impyla"):
        return cls.clientprovider.get(driver)


class ADSHiveConnection:
    def __init__(
        self,
        host: str,
        port: str = BDS_HIVE_DEFAULT_PORT,
        auth_mechanism: str = "GSSAPI",
        driver: str = "impyla",
        **kwargs,
    ):
        """Initiate the connection.

        Parameters
        ----------
        host: str
            Hive host name.
        port: str
            Hive port. Default to 10000.
        auth_mechanism: str
            Default to "GSSAPI". Using "PLAIN" for unsecure cluster.
        driver: str
            Default to "impyla". Client used to communicate with Hive. Only support impyla by far.
        kwargs:
            Other connection parameters accepted by the client.
        """
        kwargs["host"] = host
        kwargs["port"] = port
        kwargs["kerberos_service_name"] = SERVICE_NAME
        kwargs["auth_mechanism"] = auth_mechanism

        Connection = HiveConnectionFactory.get(driver)
        if not Connection:
            raise Exception(
                f"Driver {driver} does not have either required dependency or is not supported."
            )
        self.connection = Connection(**kwargs)

    def insert(
        self,
        table_name: str,
        df: pd.DataFrame,
        if_exists: str,
        batch_size: int = DEFAULT_BATCH_SIZE,
        **kwargs,
    ):
        """insert a table from a pandas dataframe.

        Parameters
        ----------
        table_name (str):
            Table Name. Table name contains database name as well. By default it will use 'default' database.
            You can specify the database name by `table_name=<db_name>.<tb_name>`.
        df (pd.DataFrame):
            Data to be injected to the database.
        if_exists (str):
            Whether to replace, append or fail if the table already exists.
        batch_size: int, default 1000
            Inserting in batches improves insertion performance. Choose this value based on available memory and network bandwidth.
        kwargs (dict):
            Other parameters used by pandas.DataFrame.to_sql.
        """
        if if_exists.lower() not in ("fail", "replace", "append"):
            raise ValueError(
                f"Unknown option `if_exists`={if_exists}. Valid options are 'fail', 'replace', 'append'."
            )

        schema = kwargs.pop("schema") if "schema" in kwargs else "default"
        engine = self.connection.get_engine(schema)
        try:
            df.to_sql(
                name=table_name.lower(),
                con=engine,
                if_exists=if_exists,
                index=False,
                method="multi",
                chunksize=batch_size,
                **kwargs,
            )
        except Exception as e:
            raise RuntimeError(f"Runtime Error: {e.args[0]}")

    def _fetch_by_batch(
        self, cursor: "impala.hiveserver2.HiveServer2Cursor", chunksize: int
    ):
        """fetch the data by batch of chunksize."""
        while True:
            rows = cursor.fetchmany(chunksize)
            if rows:
                yield rows
            else:
                break

    def query(
        self,
        sql: str,
        bind_variables: Optional[Dict] = None,
        chunksize: Optional[int] = None,
    ) -> Union["pd.DataFrame", Iterator["pd.DataFrame"]]:
        """Query data which support select statement.

        Parameters
        ----------
        sql (str):
            sql query.
        bind_variables (Optional[Dict]):
            Parameters to be bound to variables in the SQL query, if any.
            Impyla supports all DB API `paramstyle`s, including `qmark`,
            `numeric`, `named`, `format`, `pyformat`.
        chunksize (Optional[int]): . Defaults to None.
            chunksize of each of the dataframe in the iterator.

        Returns
        -------
        Union[pd.DataFrame, Iterator[pd.DataFrame]]:
            A pandas DataFrame or a pandas DataFrame iterator.
        """
        start_time = time()

        cursor = self.connection.get_cursor()

        cursor.set_arraysize(arraysize=CURSOR_SIZE)
        cursor.execute(sql, bind_variables)
        columns = [row[0] for row in cursor.description]

        if chunksize:
            logger.info(f"Chunksize is {chunksize}")
            df = iter(
                (
                    pd.DataFrame(data=rows, columns=columns)
                    for rows in self._fetch_by_batch(cursor, chunksize)
                )
            )

        else:
            df = pd.DataFrame(
                cursor,
                columns=columns,
            )
            duration = time() - start_time
            logger.info(
                f"fetched {df.shape[0]} rows at {df.shape[0]/duration:.2f} rows/seconds"
            )

        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from numbers import Integral, Number

import numpy as np

from sklearn.utils import _safe_indexing as sklearn_safe_indexing
from ads.common import logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


def _update_space_name(search_space, **kwargs):
    """
    name:
        step name
    search_space:
        search space
    """
    step_name = kwargs.pop("step_name", None)
    if step_name:
        param_distributions = {}
        for param_name, distributions in search_space.items():
            param_name = "__".join([step_name, param_name])
            param_distributions[param_name] = distributions
        return param_distributions
    else:
        return search_space


def _extract_uri(file_uri):
    bucketname, rest = file_uri.split("oci://")[1].split("@")
    namespace = rest.split("/")[0]
    filename = rest.replace(namespace + "/", "")
    return bucketname, namespace, filename


# NOTE Original implementation:
# https://github.com/scikit-learn/scikit-learn/blob/ \
# 8caa93889f85254fc3ca84caa0a24a1640eebdd1/sklearn/utils/validation.py#L131-L135
def _is_arraylike(x):
    # type: (Any) -> bool

    return hasattr(x, "__len__") or hasattr(x, "shape") or hasattr(x, "__array__")


# NOTE Original implementation:
# https://github.com/scikit-learn/scikit-learn/blob/ \
# 8caa93889f85254fc3ca84caa0a24a1640eebdd1/sklearn/utils/validation.py#L217-L234
@runtime_dependency(module="scipy", install_from=OptionalDependency.VIZ)
def _make_indexable(iterable):  # pragma: no cover
    # type: (IterableType) -> (IndexableType)

    tocsr_func = getattr(iterable, "tocsr", None)
    if tocsr_func is not None and scipy.sparse.issparse(iterable):
        return tocsr_func(iterable)
    elif (
        hasattr(iterable, "__getitem__")
        or hasattr(iterable, "iloc")
        or iterable is None
    ):
        return iterable
    return np.array(iterable)


def _num_samples(x):
    # type: (ArrayLikeType) -> int

    x_shape = getattr(x, "shape", None)
    if x_shape is not None:
        if isinstance(x_shape[0], Integral):
            return int(x_shape[0])

    try:
        return len(x)
    except TypeError:
        raise TypeError("Expected sequence or array-like, got %s." % type(x))


def _safe_indexing(
    X,  # type: Union[OneDimArrayLikeType, TwoDimArrayLikeType]
    indices,  # type: OneDimArrayLikeType
):
    # type: (...) -> Union[OneDimArrayLikeType, TwoDimArrayLikeType]
    return X if X is None else sklearn_safe_indexing(X, indices)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import pandas as pd
import warnings

from ads.common import utils, logger
from ads.dataset import helper
from ads.dataset.exception import ValidationError
from ads.dataset.dataset_with_target import ADSDatasetWithTarget
from sklearn.preprocessing import FunctionTransformer
from ads.dataset.helper import deprecate_variable, deprecate_default_value


class ClassificationDataset(ADSDatasetWithTarget):
    """
    Dataset for classification task
    """

    def __init__(self, df, sampled_df, target, target_type, shape, **kwargs):
        ADSDatasetWithTarget.__init__(
            self, df, sampled_df, target, target_type, shape, **kwargs
        )

    def auto_transform(
        self,
        fix_imbalance: bool = True,
        correlation_threshold: float = 0.7,
        frac: float = 1.0,
        correlation_methods: str = "pearson",
    ):
        """
        Return transformed dataset with several optimizations applied automatically.
        The optimizations include:

        - Dropping constant and primary key columns, which has no predictive quality,
        - Imputation, to fill in missing values in noisy data:

            - For continuous variables, fill with mean if less than 40% is missing, else drop,
            - For categorical variables, fill with most frequent if less than 40% is missing, else drop,

        - Dropping strongly co-correlated columns that tend to produce less generalizable models,
        - Balancing dataset using up or down sampling.

        Parameters
        ----------
        fix_imbalance : bool, defaults to True.
            Fix imbalance between classes in dataset. Used only for classification datasets.
        correlation_threshold: float, defaults to 0.7. It must be between 0 and 1, inclusive.
            The correlation threshold where columns with correlation higher than the threshold will
            be considered as strongly co-correlated and recommended to be taken care of.
        frac: float, defaults to 1.0. Range -> (0, 1].
            What fraction of the data should be used in the calculation?
        correlation_methods: Union[list, str], defaults to 'pearson'.

            - 'pearson': Use Pearson's Correlation between continuous features,
            - 'cramers v': Use Cramer's V correlations between categorical features,
            - 'correlation ratio': Use Correlation Ratio Correlation between categorical and continuous features,
            - 'all': Is equivalent to ['pearson', 'cramers v', 'correlation ratio'].

            Or a list containing any combination of these methods, for example, ['pearson', 'cramers v'].

        Returns
        -------
        transformed_dataset : ADSDatasetWithTarget
            The dataset after transformation

        Examples
        --------
        >>> ds_clean = ds.auto_transform(correlation_threshold=0.6)
        """
        frac = deprecate_default_value(
            frac,
            None,
            1,
            f"<code>frac=None</code> is deprecated. Use <code>frac=1.0</code> instead.",
            FutureWarning,
        )
        with utils.get_progress_bar(7) as progress:
            df, sampled_df, transformer_pipeline = self._transform(
                progress=progress,
                fix_imbalance=fix_imbalance,
                correlation_threshold=correlation_threshold,
                frac=frac,
                correlation_methods=correlation_methods,
            )
            return self._build_new_dataset(
                df,
                sampled_df=sampled_df,
                transformers=transformer_pipeline.steps,
                progress=progress,
            )

    def convert_to_text_classification(self, text_column: str):
        """
        Builds a new dataset with the given text column as the only feature besides target.

        Parameters
        ----------
        text_column: str
            Feature name to use for text classification task

        Returns
        -------
        ds: TextClassificationDataset
            Dataset with one text feature and a classification target

        Examples
        --------
        >>> review_ds = DatasetFactory.open("review_data.csv")
        >>> ds_text_class = review_ds.convert_to_text_classification('reviews')
        """

        def _select_features(df, feature_names, target):
            if target in df.columns:
                feature_names = feature_names + [target]
            return df[feature_names]

        transformer = (
            f"convert_to_text_classification using feature {text_column}",
            FunctionTransformer(
                func=_select_features,
                validate=False,
                kw_args={
                    "feature_names": [self.target.name, text_column],
                    "target": self.target.name,
                },
            ).fit(self.sampled_df),
        )
        if utils.is_same_class(self, BinaryClassificationDataset):
            new_ds = BinaryTextClassificationDataset(
                self.df[[self.target.name, text_column]],
                self.sampled_df[[self.target.name, text_column]],
                self.target.name,
                self.target.type,
                (len(self.df), 2),
                **self.init_kwargs,
            )

        else:
            new_ds = MultiClassTextClassificationDataset(
                self.df[[self.target.name, text_column]],
                self.sampled_df[[self.target.name, text_column]],
                self.target.name,
                self.target.type,
                (len(self.df), 2),
                **self.init_kwargs,
            )
        new_ds.transformer_pipeline = self._update_transformer_pipeline(transformer)
        return new_ds

    def down_sample(self, sampler=None):
        """
        Fixes an imbalanced dataset by down-sampling.

        Parameters
        ----------
        sampler: An instance of SamplerMixin
            Should implement fit_resample(X,y) method. If None, does random down sampling.

        Returns
        -------
        down_sampled_ds: ClassificationDataset
            A down-sampled dataset.

        Examples
        --------
        >>> ds = DatasetFactory.open("some_data.csv")
        >>> ds_balanced_small = ds.down_sample()
        """
        return self._build_new_dataset(
            helper.down_sample(self.df, self.target.name)
            if sampler is None
            else helper.sample(
                sampler,
                self.df.drop(self.target.name, axis=1),
                self.df[self.target.name],
            )
        )

    def up_sample(self, sampler="default"):
        """
        Fixes imbalanced dataset by up-sampling

        Parameters
        ----------
        sampler: An instance of SamplerMixin
            Should implement fit_resample(X,y) method.
            If 'default', either SMOTE or random sampler will be used
        fill_missing_type: a string
            Can either be 'mean', 'mode' or 'median'.

        Returns
        -------
        up_sampled_ds: ClassificationDataset
            an up-sampled dataset

        Examples
        --------
        >>> ds = DatasetFactory.open("some_data.csv")
        >>> ds_balanced_large = ds.up_sample()
        """
        return self._build_new_dataset(
            helper.up_sample(
                self.df,
                self.target.name,
                sampler=sampler,
                feature_types=self.feature_types,
            )
        )


class BinaryClassificationDataset(ClassificationDataset):
    """
    Dataset for binary classification
    """

    def __init__(
        self, df, sampled_df, target, target_type, shape, positive_class=None, **kwargs
    ):
        if positive_class is not None:
            # map positive_class to True
            update_arg = lambda x: x == positive_class

            def mapper(df, column_name, arg):
                df[column_name] = df[column_name].map(arg)
                return df

            df = mapper(df, target, update_arg)
            sampled_df = mapper(sampled_df, target, update_arg)
        ClassificationDataset.__init__(
            self, df, sampled_df, target, target_type, shape, **kwargs
        )

    def set_positive_class(self, positive_class, missing_value=False):
        """
        Return new dataset with values in target column mapped to True or False
        in accordance with the specified positive label.

        Parameters
        ----------
        positive_class : same dtype as target
            The target label which should be identified as positive outcome from model.
        missing_value : bool
            missing values will be converted to this

        Returns
        -------
        dataset: same type as the caller

        Raises
        ------
        ValidationError
             if the positive_class is not present in target

        Examples
        --------
        >>> ds = DatasetFactory.open("iris.csv")
        >>> ds_with_target = ds.set_target('class')
        >>> ds_with_pos_class = ds.set_positive_class('setosa')
        """
        if positive_class not in self.target.target_vals:
            raise ValidationError(
                "Positive label '%s' not in target values '%s'"
                % (positive_class, self.target.target_vals)
            )

        return self.assign_column(
            self.target.name,
            lambda x: pd.isnull(x) and missing_value or x == positive_class,
        )


class MultiClassClassificationDataset(ClassificationDataset):
    """
    Dataset for multi-class classification
    """

    def __init__(self, df, sampled_df, target, target_type, shape, **kwargs):
        ClassificationDataset.__init__(
            self, df, sampled_df, target, target_type, shape, **kwargs
        )


class BinaryTextClassificationDataset(BinaryClassificationDataset):
    """
    Dataset for binary text classification
    """

    def __init__(self, df, sampled_df, target, target_type, shape, **kwargs):
        BinaryClassificationDataset.__init__(
            self, df, sampled_df, target, target_type, shape, **kwargs
        )

    def auto_transform(self):
        """
        Automatically chooses the most effective dataset transformation
        """
        logger.info("No optimizations.")
        return self

    def select_best_features(self, score_func=None, k=12):
        """
        Automatically chooses the best features and removes the rest
        """
        logger.info(
            "There are an insufficient number of features to do feature selection."
        )
        return self


class MultiClassTextClassificationDataset(MultiClassClassificationDataset):
    """
    Dataset for multi-class text classification
    """

    def __init__(self, df, sampled_df, target, target_type, shape, **kwargs):
        MultiClassClassificationDataset.__init__(
            self, df, sampled_df, target, target_type, shape, **kwargs
        )

    def auto_transform(self):
        """
        Automatically chooses the most effective dataset transformation
        """
        logger.info("No optimizations.")
        return self

    def select_best_features(self, score_func=None, k=12):
        """
        Automatically chooses the best features and removes the rest
        """
        logger.info(
            "There are an insufficient number of features to do feature selection."
        )
        return self



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List

import pandas as pd
from ads.common.object_storage_details import ObjectStorageDetails
from ads.common.serializer import DataClassSerializable
from oci.data_labeling_service_dataplane.models.dataset import Dataset as OCIDLSDataset


@dataclass
class Metadata(DataClassSerializable):
    """The class that representing the labeled dataset metadata.

    Attributes
    ----------
    source_path: str
        Contains information on where all the source data(image/text/document) stores.
    records_path: str
        Contains information on where records jsonl file stores.
    labels: List
        List of classes/labels for the dataset.
    dataset_name: str
        Dataset display name on the Data Labeling Service console.
    compartment_id: str
        Compartment id of the labeled dataset.
    dataset_id: str
        Dataset id.
    annotation_type: str
        Type of the labeling/annotation task. Currently supports SINGLE_LABEL,
        MULTI_LABEL, ENTITY_EXTRACTION, BOUNDING_BOX.
    dataset_type: str
        Type of the dataset. Currently supports Text, Image, DOCUMENT.
    """

    source_path: str = ""
    records_path: str = ""
    labels: List[str] = field(default_factory=list)
    dataset_name: str = ""
    compartment_id: str = ""
    dataset_id: str = ""
    annotation_type: str = ""
    dataset_type: str = ""

    def to_dict(self) -> Dict:
        """Converts to dictionary representation.

        Returns
        -------
        Dict
            The metadata in dictionary type.
        """
        return asdict(self)

    def __repr__(self):
        """Show the Metadata in yaml format."""
        return self.to_yaml()

    def to_dataframe(self) -> pd.DataFrame:
        """
        Converts the metadata to dataframe format.

        Returns
        -------
        pandas.DataFrame
            The metadata in Pandas dataframe format.
        """
        return pd.DataFrame({"": self.to_dict()})

    def _repr_html_(self):
        """Shows metadata in dataframe format."""
        return (
            self.to_dataframe().style.set_properties(**{"margin-left": "0px"}).render()
        )

    @classmethod
    def from_dls_dataset(cls, dataset: OCIDLSDataset) -> "Metadata":
        """Contructs a Metadata instance from OCI DLS dataset.

        Parameters
        ----------
        dataset: OCIDLSDataset
            OCIDLSDataset object.

        Returns
        -------
        Metadata
            The ads labeled dataset metadata instance.
        """
        oci_labels = [d.name for d in dataset.label_set.items]
        oci_source_path = ObjectStorageDetails(
            dataset.dataset_source_details.bucket,
            dataset.dataset_source_details.namespace,
            dataset.dataset_source_details.prefix,
        ).path

        return Metadata(
            annotation_type=dataset.annotation_format,
            dataset_name=dataset.display_name,
            source_path=oci_source_path,
            labels=oci_labels,
            compartment_id=dataset.compartment_id,
            dataset_id=dataset.id,
            dataset_type=dataset.dataset_format_details.format_type,
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""This module runs a Jupyter Python notebook with nbconvert and print the outputs.
This is a driver script auto-generated by Oracle ADS.

The following environment variables are used:
JOB_RUN_NOTEBOOK:
    The relative path of the jupyter Python notebook to be executed.
NOTEBOOK_EXCLUDE_TAGS:
    Optional, a list of tags serialized to JSON string.
    Notebook cells with one of the tags will be excluded from running.
NOTEBOOK_ENCODING:
    Optional, the encoding for opening the notebook.
OUTPUT_URI:
    Optional, object storage URI for saving files from the output directory.
"""
import logging
import json
import os
from typing import Optional

import nbformat
from nbconvert.preprocessors import ExecutePreprocessor, CellExecutionError

try:
    # This is used by ADS and testing
    from .driver_utils import OCIHelper, JobRunner, set_log_level
except ImportError:
    # This is used when the script is in a job run.
    from driver_utils import OCIHelper, JobRunner, set_log_level


logger = logging.getLogger(__name__)
logger = set_log_level(logger)

class ADSExecutePreprocessor(ExecutePreprocessor):
    """Customized Execute Preprocessor for running notebook."""

    def __init__(self, exclude_tags=None, **kw):
        """Initialize the preprocessor

        Parameters
        ----------
        exclude_tags : list, optional
            A list of cell tags, notebook cells with any of these cell tag will be skipped.
            Defaults to None.
        """
        self.exclude_tags = exclude_tags
        super().__init__(**kw)

    @staticmethod
    def _print_cell_outputs(cell):
        """Prints the outputs of a notebook cell."""
        for output in cell.outputs:
            output_type = output.get("output_type")
            if output_type == "stream":
                # stream outputs includes line break already
                print(output.text, end="")
            elif output_type == "execute_result":
                # execute_result may contain text/plain
                text = output.get("data", {}).get("text/plain", [])
                # The value could be str or list of str
                if isinstance(text, list):
                    for line in text:
                        print(line)
                else:
                    print(text)

    def preprocess_cell(self, cell, resources, *args, **kwargs):
        """Runs the notebook cell and print out the outputs"""
        # Skip the cell if any of the cell tags matching an exclude tag.
        if self.exclude_tags:
            # Log an error message if there is an error reading the cell tags,
            # and continue to run the cell.
            try:
                cell_tags = cell.get("metadata", {}).get("tags", [])
                for tag in cell_tags:
                    if tag in self.exclude_tags:
                        return cell, resources
            except Exception as ex:
                logger.exception("An error occurred when reading cell tags.")
        # Run the cell
        cell, resources = super().preprocess_cell(cell, resources, *args, **kwargs)
        # Print cell output
        if hasattr(cell, "outputs"):
            # Log a message if there is an error getting the cell output,
            # and continue to run the next cell.
            try:
                self._print_cell_outputs(cell)
            except Exception as ex:
                logger.exception("An error occurred when reading cell outputs.")
        return cell, resources


def run_notebook(
    notebook_path: str,
    working_dir: Optional[str] = None,
    exclude_tags: Optional[list] = None,
) -> Optional[CellExecutionError]:
    """Runs a notebook

    Parameters
    ----------
    notebook_path : str
        The path of the notebook
    working_dir : str, optional
        The working directory for running the notebook, by default None.
        If this is None, the same directory of the notebook_path will be used.
    exclude_tags : list, optional
        Tags for excluding cells, by default None

    Returns
    -------
    CellExecutionError or None
        Exception object when there is an error in a notebook cell.
        Otherwise, None.
    """
    # Read the notebook
    encoding = os.environ.get("NOTEBOOK_ENCODING", "utf-8")
    with open(notebook_path, encoding=encoding) as f:
        nb = nbformat.read(f, as_version=4)

    # Working/Output directory
    if not working_dir:
        working_dir = os.path.dirname(notebook_path)

    # The path of the output notebook with results/plots
    notebook_filename_out = os.path.join(working_dir, os.path.basename(notebook_path))

    ep = ADSExecutePreprocessor(exclude_tags=exclude_tags, kernel_name="python")

    try:
        ep.preprocess(nb, {"metadata": {"path": working_dir}})
        ex = None
    except CellExecutionError as exc:
        msg = "Error executing the notebook.\n\n"
        logger.error(msg)
        ex = exc
    finally:
        with open(notebook_filename_out, mode="w", encoding=encoding) as f:
            nbformat.write(nb, f)
    return ex


def main() -> None:
    """Runs the driver to execute a notebook."""
    JobRunner().conda_unpack()

    notebook_file_path = os.path.join(
        os.path.dirname(__file__), os.environ.get("JOB_RUN_NOTEBOOK")
    )
    output_dir = os.path.join(os.path.dirname(__file__), "outputs")
    # Create the output directory
    os.makedirs(output_dir, exist_ok=True)
    # Exclude tags
    tags = os.environ.get("NOTEBOOK_EXCLUDE_TAGS")
    if tags:
        tags = json.loads(tags)
        logger.info("Excluding cells with any of the following tags: %s", tags)
    # Run the notebook
    ex = run_notebook(notebook_file_path, working_dir=output_dir, exclude_tags=tags)

    # Save the outputs
    OCIHelper.copy_outputs(output_dir)

    if ex:
        raise ex


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a LatLong feature type.

Classes:
    LatLong
        The LatLong feature type.

Functions:
    default_handler(data: pd.Series) -> pd.Series
        Processes given data and indicates if the data matches requirements.
"""
import matplotlib.pyplot as plt
import pandas as pd
import re
from ads.feature_engineering.feature_type.string import String
from ads.feature_engineering.utils import (
    _count_unique_missing,
    _str_lat_long_to_point,
    SchemeNeutral,
    SchemeTeal,
)
from ads.feature_engineering import schema
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


PATTERN = re.compile(r"^[(]?(\-?\d+\.\d+?),\s*(\-?\d+\.\d+?)[)]?$", re.VERBOSE)


def default_handler(data: pd.Series, *args, **kwargs) -> pd.Series:
    """Processes given data and indicates if the data matches requirements.

    Parameters
    ----------
    data: :class:`pandas.Series`
        The data to process.

    Returns
    -------
    :class:`pandas.Series`
        The logical list indicating if the data matches requirements.
    """
    return data.apply(
        lambda x: True
        if not pd.isnull(x) and PATTERN.match(str(x)) is not None
        else False
    )


class LatLong(String):
    """Type representing longitude and latitute.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Shows the location of given address on map base on longitude and latitute.

    Example
    -------
    >>> from ads.feature_engineering.feature_type.lat_long import LatLong
    >>> import pandas as pd
    >>> s = pd.Series(["-18.2193965, -93.587285",
                        "-21.0255305, -122.478584",
                        "85.103913, 19.405744",
                        "82.913736, 178.225672",
                        "62.9795085,-66.989705",
                        "54.5604395,95.235090",
                        "24.2811855,-162.380403",
                        "-1.818319,-80.681214",
                        None,
                        "(51.816119, 175.979008)",
                        "(54.3392995,-11.801615)"],
                        name='latlong')
    >>> s.ads.feature_type = ['lat_long']
    >>> LatLong.validator.is_lat_long(s)
    0      True
    1      True
    2      True
    3      True
    4      True
    5      True
    6      True
    7      True
    8     False
    9      True
    10     True
    Name: latlong, dtype: bool
    """

    description = "Type representing longitude and latitute."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generate feature statistics.

        Feature statistics include (total)count, unique(count)
        and missing(count) if there is any.

        Examples
        --------
        >>> latlong_series = pd.Series([
                    "69.196241,-125.017615",
                    "5.2272595,-143.465712",
                    "-33.9855425,-153.445155",
                    "43.340610,86.460554",
                    "24.2811855,-162.380403",
                    "2.7849025,-7.328156",
                    "45.033805,157.490179",
                    "-1.818319,-80.681214",
                    "-44.510428,-169.269477",
                    "-56.3344375,-166.407038",
                    "",
                    np.NaN,
                    None
                ],
            name='latlong'
        )
        >>> latlong_series.ads.feature_type = ['lat_long']
        >>> latlong_series.ads.feature_stat()
            Metric  Value
        0	count	13
        1	unique	10
        2	missing	3

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series or Dataframe provided.
        """
        return _count_unique_missing(x)

    @staticmethod
    @runtime_dependency(module="geopandas", install_from=OptionalDependency.GEO)
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows the location of given address on map base on longitude and latitute.

        Examples
        --------
        >>> latlong_series = pd.Series([
                    "69.196241,-125.017615",
                    "5.2272595,-143.465712",
                    "-33.9855425,-153.445155",
                    "43.340610,86.460554",
                    "24.2811855,-162.380403",
                    "2.7849025,-7.328156",
                    "45.033805,157.490179",
                    "-1.818319,-80.681214",
                    "-44.510428,-169.269477",
                    "-56.3344375,-166.407038",
                    "",
                    np.NaN,
                    None
                ],
            name='latlong'
        )
        >>> latlong_series.ads.feature_type = ['lat_long']
        >>> latlong_series.ads.feature_plot()

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the LatLong feature type.
        """
        col_name = x.name if x.name else "latlong"
        df = x.to_frame(name=col_name)
        df["validation"] = default_handler(x)
        df = df[df["validation"] == True]
        df = (
            df[col_name]
            .apply(lambda x: _str_lat_long_to_point(x))
            .to_frame(name="Coordinates")
            .dropna()
        )
        if len(df.index):
            df["Coordinates"] = geopandas.GeoSeries.from_wkt(df["Coordinates"])
            gdf = geopandas.GeoDataFrame(df, geometry="Coordinates")
            fig, ax = plt.subplots(facecolor=SchemeNeutral.BACKGROUND_LIGHT)
            world = geopandas.read_file(
                geopandas.datasets.get_path("naturalearth_lowres")
            )
            world.plot(
                ax=ax, color=SchemeNeutral.AREA_LIGHT, linewidth=0.5, edgecolor="white"
            )
            gdf.plot(ax=ax, color=SchemeTeal.LINE_LIGHT, markersize=10)
            plt.axis("off")
            return ax

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> latlong_series = pd.Series([
            "69.196241,-125.017615",
            "5.2272595,-143.465712",
            "-33.9855425,-153.445155",
            "43.340610,86.460554",
            "24.2811855,-162.380403",
            "2.7849025,-7.328156",
            "45.033805,157.490179",
            "-1.818319,-80.681214",
            "-44.510428,-169.269477",
            "-56.3344375,-166.407038",
            "",
            np.NaN,
            None
            ],
            name='latlong'
        )
        >>> latlong_series.ads.feature_type = ['lat_long']
        >>> latlong_series.ads.feature_domain()
        constraints: []
        stats:
            count: 13
            missing: 3
            unique: 10
        values: LatLong

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the LatLong feature type.
        """

        return schema.Domain(
            cls.__name__,
            cls.feature_stat(x).to_dict()[x.name],
            [],
        )


LatLong.validator.register("is_lat_long", default_handler)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

from sklearn.base import TransformerMixin

from ads.dataset.progress import DummyProgressBar


class FeatureEngineeringTransformer(TransformerMixin):
    def __init__(self, feature_metadata=None):
        self.feature_metadata_ = feature_metadata
        self.function_ = None
        self.function_kwargs_ = None

    def __repr__(self):
        return "No feature engineering transformations"

    def fit(self, X, y=None):
        self.function_ = None
        self.function_kwargs_ = None
        del self.feature_metadata_
        return self

    def fit_transform(self, X, y=None, **fit_params):
        return self.fit(X, y=y).transform(X, fit_transform=True)

    def transform(self, df, progress=DummyProgressBar(), fit_transform=False):
        if self.function_ is not None:
            return df.pipe(self.function_, **self.function_kwargs_)
        return df



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import logging

logger = logging.getLogger(__name__)
try:
    from ads.jobs.builders.runtimes.python_runtime import (
        PythonRuntime,
        GitPythonRuntime,
        NotebookRuntime,
        ScriptRuntime,
        DataFlowRuntime,
        DataFlowNotebookRuntime,
    )
    from ads.jobs.builders.runtimes.container_runtime import ContainerRuntime
    from ads.jobs.ads_job import Job
    from ads.jobs.builders import infrastructure
    from ads.jobs.builders.infrastructure.dsc_job import (
        DataScienceJob,
        DataScienceJobRun,
    )
    from ads.jobs.builders.infrastructure.dataflow import DataFlow, DataFlowRun
except AttributeError as e:
    import oci.data_science

    if not hasattr(oci.data_science.models, "Job"):
        logger.warning(
            "The OCI SDK you installed does not support Data Science Jobs. ADS Jobs API will not work."
        )
    else:
        raise e

__all__ = [
    "Job",
    "DataScienceJob",
    "DataScienceJobRun",
    "PythonRuntime",
    "GitPythonRuntime",
    "NotebookRuntime",
    "ScriptRuntime",
    "ContainerRuntime",
    "DataFlow",
    "DataFlowRun",
    "DataFlowRuntime",
    "DataFlowNotebookRuntime",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import, division

import copy
import datetime
import fsspec
import numpy as np
import os
import pandas as pd
import uuid

from collections import Counter
from sklearn.preprocessing import FunctionTransformer
from typing import Iterable, Union

from ads import set_documentation_mode
from ads.common import utils
from ads.common.decorator.deprecate import deprecated
from ads.dataset import helper, logger
from ads.dataset.dataframe_transformer import DataFrameTransformer
from ads.dataset.exception import ValidationError
from ads.dataset.helper import (
    convert_columns,
    fix_column_names,
    generate_sample,
    DatasetDefaults,
    deprecate_default_value,
    deprecate_variable,
)
from ads.dataset.label_encoder import DataFrameLabelEncoder
from ads.dataset.pipeline import TransformerPipeline
from ads.dataset.progress import DummyProgressBar
from ads.dataset.sampled_dataset import PandasDataset
from ads.type_discovery.type_discovery_driver import TypeDiscoveryDriver
from ads.dataset.helper import get_feature_type
from ads.dataset.correlation_plot import plot_correlation_heatmap
from ads.dataset.correlation import (
    _cat_vs_cts,
    _cat_vs_cat,
    _get_columns_by_type,
    _validate_correlation_methods,
)
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)

N_Features_Wide_Dataset = 64


pd.set_option("display.max_colwidth", None)


class ADSDataset(PandasDataset):
    """
    An ADSDataset Object.

    The ADSDataset object cannot be used for classification or regression problems until a
    target has been set using `set_target`. To see some rows in the data use any of the usual
    Pandas functions like `head()`. There are also a variety of converters, to_dask,
    to_pandas, to_h2o, to_xgb, to_csv, to_parquet, to_json & to_hdf .
    """

    df_read_functions = ["head", "describe", "_get_numeric_data"]

    def __init__(
        self,
        df,
        sampled_df,
        shape,
        name="",
        description=None,
        type_discovery=True,
        types={},
        metadata=None,
        progress=DummyProgressBar(),
        transformer_pipeline=None,
        interactive=False,
        **kwargs,
    ):

        #
        # to keep performance high and linear no matter the size of the distributed dataset we
        # create a pandas df that's used internally because this has a fixed upper size.
        #
        super().__init__(
            sampled_df,
            type_discovery=type_discovery,
            types=types,
            metadata=metadata,
            progress=progress,
        )
        self.df = fix_column_names(df)

        self.name = name
        self.description = description
        self.shape = shape
        # store these args to reapply when building a new dataset for delegate operations on dataframe
        self.init_kwargs = {**kwargs, "type_discovery": type_discovery}
        if transformer_pipeline is None:
            # Update transformer pipeline to convert column types and fix names
            self.transformer_pipeline = TransformerPipeline(
                steps=[
                    (
                        "prepare",
                        FunctionTransformer(func=fix_column_names, validate=False),
                    )
                ]
            )
            self.transformer_pipeline = self._update_transformer_pipeline(
                steps=[
                    (
                        "type_discovery",
                        FunctionTransformer(
                            func=convert_columns,
                            validate=False,
                            kw_args={"dtypes": self.sampled_df.dtypes},
                        ),
                    )
                ]
            )
        else:
            self.transformer_pipeline = transformer_pipeline

    def __repr__(self):
        rows, cols = self.shape
        return f"{rows:,} rows, {cols:,} columns"

    def __len__(self):
        return self.shape[0]

    @property
    @deprecated(
        "2.5.2", details="The ddf attribute is deprecated. Use the df attribute."
    )
    def ddf(self):
        return self.df

    @deprecated(
        "2.5.2", details="The compute method is deprecated. Use the df attribute."
    )
    def compute(self):
        return self.df

    @runtime_dependency(
        module="ipywidgets", object="HTML", install_from=OptionalDependency.NOTEBOOK
    )
    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def _repr_html_(self):
        from IPython.core.display import display, HTML

        display(
            HTML(
                utils.horizontal_scrollable_div(
                    self.sampled_df.head(5)
                    .style.set_table_styles(utils.get_dataframe_styles())
                    .set_table_attributes("class=table")
                    .hide_index()
                    .render()
                )
            )
        )

    def _head(self, n=5):
        """
        Return the first `n` rows of the dataset.

        Parameters
        ----------
        n : int, default 5
            Number of rows to select.

        Returns
        -------
        dataset_head : pandas.DataFrame
            The first `n` rows of the dataset

        Examples
        --------
        >>> ds = DatasetFactory.open("classfication_data.csv")
        >>> ds.head()
        * displays the first 5 rows of the dataset, just as the traditional head() function would *
        """
        df = self.df.head(n=n)

        #
        # we could just return the above but, jupyterlab doesn't render these well
        # when the width exceeds the screen area. To address that we wrap the dataframe
        # with a class that has an optimized _repr_html_ handler, this object
        # extends the pandas dataframe so it can still be used as-a dataframe
        #
        class FormattedDataFrame(pd.DataFrame):
            def __init__(self, *args, **kwargs):
                super(FormattedDataFrame, self).__init__(*args, **kwargs)

            @property
            def _constructor(self):
                return FormattedDataFrame

            @runtime_dependency(
                module="ipywidgets",
                object="HTML",
                install_from=OptionalDependency.NOTEBOOK,
            )
            @runtime_dependency(
                module="IPython", install_from=OptionalDependency.NOTEBOOK
            )
            def _repr_html_(self):
                from IPython.core.display import display, HTML

                display(
                    HTML(
                        utils.horizontal_scrollable_div(
                            self.style.set_table_styles(utils.get_dataframe_styles())
                            .set_table_attributes("class=table")
                            .hide_index()
                            .render()
                        )
                    )
                )
                return None

            def __repr__(self):
                return "{} rows, {} columns".format(*self.shape)

        return FormattedDataFrame(df)

    def call(self, func, *args, sample_size=None, **kwargs):
        r"""
        Runs a custom function on dataframe

        func will receive the pandas dataframe (which represents the dataset) as an argument named 'df' by default.
        This can be overridden by specifying the dataframe argument name in a tuple (func, dataframe_name).

        Parameters
        ----------
        func: Union[callable, tuple]
            Custom function that takes pandas dataframe as input
            Alternatively a (callable, data) tuple where data is a string indicating the keyword of callable
            that expects the dataframe name
        args: iterable, optional
            Positional arguments passed into func
        sample_size: int, Optional
            To use a sampled dataframe
        kwargs: mapping, optional
            A dictionary of keyword arguments passed into func

        Returns
        -------
        func: function
            a plotting function that contains `*args` and `**kwargs`

        Examples
        --------
        >>> ds = DatasetFactory.open("classfication_data.csv")
        >>> def f1(df):
        ...  return(sum(df), axis=0)
        >>> sum_ds = ds.call(f1)
        """

        data = "df"
        if isinstance(func, tuple):
            func, data = func
            if data in kwargs:
                raise ValueError(
                    "'%s' is both the data argument and a keyword argument" % data
                )

        if sample_size is None:
            # user has asked not to do sampling
            df = self.df.copy()
        else:
            df = self.df.sample(n=sample_size)
        kwargs[data] = df
        return func(*args, **kwargs)

    def set_target(self, target, type_discovery=True, target_type=None):
        """
        Returns a dataset tagged based on the type of target.

        Parameters
        ----------
        target: str
            name of the feature to use as target.
        type_discovery: bool
            This is set as True by default.
        target_type: type
            If provided, then the target will be typed with the provided value.

        Returns
        -------
        ds: ADSDataset
            tagged according to the type of the target column.

        Examples
        --------
        >>> ds = DatasetFactory.open("classfication_data.csv")
        >>> ds_with_target= ds.set_target("target_class")
        """
        from ads.dataset.factory import DatasetFactory

        if target_type:
            target_series = self.sampled_df[target].astype(target_type)
        else:
            target_series = self.sampled_df[target]
        return DatasetFactory._get_dataset(
            self.df,
            self.sampled_df,
            target,
            DatasetFactory.infer_target_type(target, target_series, type_discovery),
            self.shape,
            **self.init_kwargs,
        )

    @deprecated("2.5.2", details="Instead use `to_pandas`.")
    def to_pandas_dataframe(
        self, filter=None, frac=None, include_transformer_pipeline=False
    ):
        return self.to_pandas(
            filter=filter,
            frac=frac,
            include_transformer_pipeline=include_transformer_pipeline,
        )

    def to_pandas(self, filter=None, frac=None, include_transformer_pipeline=False):
        """
        Returns a copy of the data as pandas.DataFrame, and a sklearn pipeline optionally that holds the
        transformations run so far on the data.

        The pipeline returned can be updated with the transformations done offline and passed along with the
        dataframe to Dataset.open API if the transformations need to be reproduced at the time of scoring.

        Parameters
        ----------
        filter: str, optional
            The query string to filter the dataframe, for example
            ds.to_pandas(filter="age > 50 and location == 'san francisco")
            See also https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.query.html
        frac: float, optional
            fraction of original data to return.
        include_transformer_pipeline: bool, default: False
            If True, (dataframe, transformer_pipeline) is returned as a tuple

        Returns
        -------
        dataframe : pandas.DataFrame
            if include_transformer_pipeline is False.
        (data, transformer_pipeline): tuple of pandas.DataFrame and dataset.pipeline.TransformerPipeline
            if include_transformer_pipeline is True.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds_as_df = ds.to_pandas()

        Notes
        -----
        See also https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html#sklearn.pipeline.Pipeline
        """
        df = self.df.query(filter) if filter is not None else self.df.copy()
        if frac is not None:
            df = df.sample(frac=frac)
        return (
            (df, copy.deepcopy(self.transformer_pipeline))
            if include_transformer_pipeline
            else df
        )

    @deprecated("2.5.2", details="Instead use `to_dask`.")
    def to_dask_dataframe(
        self,
        filter=None,
        frac=None,
        npartitions=None,
        include_transformer_pipeline=False,
    ):
        return self.to_dask(
            filter=filter,
            frac=frac,
            npartitions=npartitions,
            include_transformer_pipeline=include_transformer_pipeline,
        )

    @runtime_dependency(module="dask.dataframe", short_name="dd")
    def to_dask(
        self,
        filter=None,
        frac=None,
        npartitions=None,
        include_transformer_pipeline=False,
    ):
        """
        Returns a copy of the data as dask.dataframe.core.DataFrame, and a sklearn pipeline optionally that holds the
        transformations run so far on the data.

        The pipeline returned can be updated with the transformations done offline and passed along with the
        dataframe to Dataset.open API if the transformations need to be reproduced at the time of scoring.

        Parameters
        ----------
        filter: str, optional
            The query string to filter the dataframe, for example
            ds.to_dask(filter="age > 50 and location == 'san francisco")
            See also https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.query.html
        frac: float, optional
            fraction of original data to return.
        include_transformer_pipeline: bool, default: False
            If True, (dataframe, transformer_pipeline) is returned as a tuple.

        Returns
        -------
        dataframe : dask.dataframe.core.DataFrame
            if include_transformer_pipeline is False.
        (data, transformer_pipeline): tuple of dask.dataframe.core.DataFrame and dataset.pipeline.TransformerPipeline
            if include_transformer_pipeline is True.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds_dask = ds.to_dask()

        Notes
        -----
        See also http://docs.dask.org/en/latest/dataframe-api.html#dataframe and
        https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html#sklearn.pipeline.Pipeline

        """
        res = self.to_pandas(
            filter=filter,
            frac=frac,
            include_transformer_pipeline=include_transformer_pipeline,
        )
        return (
            (dd.from_pandas(res[0], npartitions=npartitions), res[1])
            if include_transformer_pipeline
            else dd.from_pandas(res, npartitions=npartitions)
        )

    @deprecated("2.5.2", details="Instead use `to_h2o`.")
    def to_h2o_dataframe(
        self, filter=None, frac=None, include_transformer_pipeline=False
    ):
        return self.to_h2o(
            filter=filter,
            frac=frac,
            include_transformer_pipeline=include_transformer_pipeline,
        )

    @runtime_dependency(module="h2o")
    def to_h2o(self, filter=None, frac=None, include_transformer_pipeline=False):
        """
        Returns a copy of the data as h2o.H2OFrame, and a sklearn pipeline optionally that holds the
        transformations run so far on the data.

        The pipeline returned can be updated with the transformations done offline and passed along with the
        dataframe to Dataset.open API if the transformations need to be reproduced at the time of scoring.

        Parameters
        ----------
        filter: str, optional
            The query string to filter the dataframe, for example
            ds.to_h2o(filter="age > 50 and location == 'san francisco")
            See also https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.query.html
        frac: float, optional
            fraction of original data to return.
        include_transformer_pipeline: bool, default: False
            If True, (dataframe, transformer_pipeline) is returned as a tuple.

        Returns
        -------
        dataframe : h2o.H2OFrame
            if include_transformer_pipeline is False.
        (data, transformer_pipeline): tuple of  h2o.H2OFrame and dataset.pipeline.TransformerPipeline
            if include_transformer_pipeline is True.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds_as_h2o = ds.to_h2o()

        Notes
        -----
        See also https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html#sklearn.pipeline.Pipeline
        """
        res = self.to_pandas(
            filter=filter,
            frac=frac,
            include_transformer_pipeline=include_transformer_pipeline,
        )
        return (
            (h2o.H2OFrame(res[0]), res[1])
            if include_transformer_pipeline
            else h2o.H2OFrame(res)
        )

    @deprecated("2.5.2", details="Instead use `to_xgb`.")
    def to_xgb_dmatrix(
        self, filter=None, frac=None, include_transformer_pipeline=False
    ):
        return self.to_xgb(
            filter=filter,
            frac=frac,
            include_transformer_pipeline=include_transformer_pipeline,
        )

    @runtime_dependency(module="xgboost", install_from=OptionalDependency.BOOSTED)
    def to_xgb(self, filter=None, frac=None, include_transformer_pipeline=False):
        """
        Returns a copy of the data as xgboost.DMatrix, and a sklearn pipeline optionally that holds the
        transformations run so far on the data.

        The pipeline returned can be updated with the transformations done offline and passed along with the
        dataframe to Dataset.open API if the transformations need to be reproduced at the time of scoring.

        Parameters
        ----------
        filter: str, optional
            The query string to filter the dataframe, for example
            ds.to_xgb(filter="age > 50 and location == 'san francisco")
            See also https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.query.html
        frac: float, optional
            fraction of original data to return.
        include_transformer_pipeline: bool, default: False
            If True, (dataframe, transformer_pipeline) is returned as a tuple.

        Returns
        -------
        dataframe : xgboost.DMatrix
            if include_transformer_pipeline is False.
        (data, transformer_pipeline): tuple of xgboost.DMatrix and dataset.pipeline.TransformerPipeline
            if include_transformer_pipeline is True.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> xgb_dmat = ds.to_xgb()

        Notes
        -----
        See also https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html#sklearn.pipeline.Pipeline
        """
        res = self.to_pandas(
            filter=filter,
            frac=frac,
            include_transformer_pipeline=include_transformer_pipeline,
        )
        df = res[0] if include_transformer_pipeline else res
        le = DataFrameLabelEncoder()
        df = le.fit_transform(df)
        if include_transformer_pipeline:
            res[1].add(le)
        xgb_matrix = xgboost.DMatrix(df)
        return (xgb_matrix, res[1]) if include_transformer_pipeline else xgb_matrix

    def sample(self, frac=None, random_state=utils.random_state):
        """
        Returns random sample of dataset.

        Parameters
        ----------
        frac : float, optional
            Fraction of axis items to return.
        random_state : int or ``np.random.RandomState``
            If int we create a new RandomState with this as the seed
            Otherwise we draw from the passed RandomState

        Returns
        -------
        sampled_dataset: ADSDataset
            An ADSDataset which was randomly sampled.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds_sample = ds.sample()
        """
        df = self.df.sample(frac=frac, random_state=random_state)
        return self._build_new_dataset(df)

    def drop_columns(self, columns):
        """
        Return new dataset with specified columns removed.

        Parameters
        ----------
        columns : str or list
            columns to drop.

        Returns
        -------
        dataset: same type as the caller
            a dataset with specified columns dropped.

        Raises
        ------
        ValidationError
            If any of the feature names is not found in the dataset.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds_smaller = ds.drop_columns(['col1', 'col2'])
        """
        self._validate_feature(columns)
        return self.drop(columns, axis=1)

    def assign_column(self, column, arg):
        """
        Return new dataset with new column or values of the existing column mapped according to input correspondence.

        Used for adding a new column or substituting each value in a column with another value, that may be derived from
        a function, a :class:`pandas.Series` or a :class:`pandas.DataFrame`.

        Parameters
        ----------
        column : str
            Name of the feature to update.
        arg : function, dict, Series or DataFrame
            Mapping correspondence.

        Returns
        -------
        dataset: same type as the caller
            a dataset with the specified column assigned.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds_same_size = ds.assign_column('target',lambda x:  x>15 if x not None)
        >>> ds_bigger = ds.assign_column('new_col', np.arange(ds.shape[0]))
        """
        target_name = (
            self.target.name if not utils.is_same_class(self, ADSDataset) else None
        )
        if isinstance(arg, Iterable) or isinstance(arg, ADSDataset):
            df = self.df.copy()
            if type(arg) == pd.DataFrame:
                col_to_add = arg
            elif type(arg) == ADSDataset:
                col_to_add = arg.df
            elif type(arg) == dict:
                col_to_add = pd.DataFrame.from_dict(arg)
            elif type(arg) in [list, np.ndarray]:
                col_to_add = pd.DataFrame(arg, columns=["new_col"])
            elif type(arg) == pd.Series:
                col_to_add = arg.rename("new_col").to_frame()
            elif utils._is_dask_dataframe(arg):
                col_to_add = arg.compute()
            elif utils._is_dask_series(arg):
                col_to_add = arg.compute().rename("new_col").to_frame()
            else:
                raise ValueError(
                    f"assign_column currently does not support arg of type {type(arg)}. Reformat "
                    f"as types: Pandas, numpy, list, or dict"
                )
            if column in df.columns:
                df = df.drop(columns=column)
            new_df = pd.concat([df, col_to_add], axis=1).rename(
                columns={"new_col": column}
            )
            return self._build_new_dataset(new_df)

        else:
            sampled_df = self.sampled_df.copy()
            df = self.df.copy()
            sampled_df[column] = sampled_df[column].apply(arg)
            df[column] = df[column].apply(arg)
            if column == target_name:
                target_type = get_feature_type(target_name, sampled_df[target_name])
                return self._build_new_dataset(
                    df, sampled_df, target=target_name, target_type=target_type
                )
            else:
                return self._build_new_dataset(
                    df,
                    sampled_df,
                    target=target_name,
                    target_type=self.target.type
                    if target_name != column and target_name is not None
                    else None,
                )

    def rename_columns(self, columns):
        """
        Returns a new dataset with altered column names.

        dict values must be unique (1-to-1). Labels not contained in a dict will be left as-is.
        Extra labels listed don't throw an error.

        Parameters
        ----------
        columns: dict-like or function or list of str
            dict to rename columns selectively, or list of names to rename all columns, or a function like
            str.upper

        Returns
        -------
        dataset: same type as the caller
            A dataset with specified columns renamed.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds_renamed = ds.rename_columns({'col1': 'target'})
        """
        if isinstance(columns, list):
            assert len(columns) == len(
                self.columns.values
            ), "columns length do not match the dataset"
            columns = dict(zip(self.columns.values, columns))
        return self.rename(columns=columns)

    def set_name(self, name):
        """
        Sets name for the dataset.

        This name will be used to filter the datasets returned by ds.list() API.
        Calling this API is optional. By default name of the dataset is set to empty.

        Parameters
        ----------
        name: str
            Name of the dataset.

        Examples
        --------
        >>> ds = DatasetFactory.open("data1.csv")
        >>> ds_renamed = ds.set_name("dataset1")
        """
        self.name = name

    def set_description(self, description):
        """
        Sets description for the dataset.

        Give your dataset a description.

        Parameters
        ----------
        description: str
            Description of the dataset.

        Examples
        --------
        >>> ds = DatasetFactory.open("data1.csv")
        >>> ds_renamed = ds.set_description("dataset1 is from "data1.csv"")
        """
        self.description = description

    def snapshot(self, snapshot_dir=None, name="", storage_options=None):
        """
        Snapshot the dataset with modifications made so far.

        Optionally  caller can invoke ds.set_name() before saving to identify the dataset uniquely at the time of
        using ds.list().

        The snapshot can be reloaded by providing the URI returned by this API to DatasetFactory.open()

        Parameters
        ----------
        snapshot_dir: str, optional
            Directory path under which dataset snapshot will be created.
            Defaults to snapshots_dir set using DatasetFactory.set_default_storage().
        name: str, optional, default: ""
            Name to uniquely identify the snapshot using DatasetFactory.list_snapshots().
            If not provided, an auto-generated name is used.
        storage_options: dict, optional
            Parameters passed on to the backend filesystem class.
            Defaults to storage_options set using DatasetFactory.set_default_storage().

        Returns
        -------
        p_str: str
            the URI to access the snapshotted dataset.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds_uri = ds.snapshot()
        """
        if snapshot_dir is None:
            import ads.dataset.factory as factory

            snapshot_dir = factory.default_snapshots_dir
            if snapshot_dir is None:
                raise ValueError(
                    "Specify snapshot_dir or use DatasetFactory.set_default_storage() to set default \
                                  storage options"
                )
            else:
                logger.info("Using default snapshots dir %s" % snapshot_dir)
        name = self._get_unique_name(name)
        if not snapshot_dir.endswith("/"):
            snapshot_dir = snapshot_dir + "/"
        parquet_file = "%s%s.parquet" % (snapshot_dir, name)
        os.makedirs(snapshot_dir, exist_ok=True)
        if storage_options is None and parquet_file[:3] == "oci":
            import ads.dataset.factory as factory

            storage_options = factory.default_storage_options
            logger.info("Using default storage options.")

        return helper.write_parquet(
            path=parquet_file,
            data=self.df,
            metadata_dict={
                "metadata": self.feature_types,
                "transformer": self.transformer_pipeline,
            },
            storage_options=storage_options,
        )

    def to_csv(self, path, storage_options=None, **kwargs):
        """
        Save the materialized dataframe to csv file.

        Parameters
        ----------
        path: str
            Location to write to. If there are more than one partitions in df, should include a glob character to
            expand into a set of file names, or provide a `name_function=parameter`.
            Supports protocol specifications such as `"oci://"`, `"s3://"`.
        storage_options: dict, optional
             Parameters passed on to the backend filesystem class.
             Defaults to storage_options set using DatasetFactory.set_default_storage().
        kwargs: dict, optional

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> [ds_link] = ds.to_csv("my/path.csv")
        """
        if storage_options is None:
            import ads.dataset.factory as factory

            storage_options = factory.default_storage_options
            logger.info("Using default storage options")
        return self.df.to_csv(path, storage_options=storage_options, **kwargs)

    def to_parquet(self, path, storage_options=None, **kwargs):
        """
        Save data to parquet file.

        Parameters
        ----------
        path: str
            Location to write to. If there are more than one partitions in df, should include a glob character to
            expand into a set of file names, or provide a `name_function=parameter`.
            Supports protocol specifications such as `"oci://"`, `"s3://"`.
        storage_options: dict, optional
             Parameters passed on to the backend filesystem class.
             Defaults to storage_options set using DatasetFactory.set_default_storage().
        kwargs: dict, optional

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds.to_parquet("my/path")
        """
        if storage_options is None:
            import ads.dataset.factory as factory

            storage_options = factory.default_storage_options
            logger.info("Using default storage options")
        return self.df.to_parquet(path, storage_options=storage_options, **kwargs)

    def to_json(self, path, storage_options=None, **kwargs):
        """
        Save data to JSON files.

        Parameters
        ----------
        path: str
            Location to write to. If there are more than one partitions in df, should include a glob character to
            expand into a set of file names, or provide a `name_function=parameter`.
            Supports protocol specifications such as `"oci://"`, `"s3://"`.
        storage_options: dict, optional
            Parameters passed on to the backend filesystem class.
            Defaults to storage_options set using DatasetFactory.set_default_storage().
        kwargs: dict, optional

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds.to_json("my/path.json")
        """
        if storage_options is None:
            import ads.dataset.factory as factory

            storage_options = factory.default_storage_options
            logger.info("Using default storage options")

        return self.df.to_json(path, storage_options=storage_options, **kwargs)

    def to_hdf(
        self, path: str, key: str, storage_options: dict = None, **kwargs
    ) -> str:
        """
        Save data to Hierarchical Data Format (HDF) files.

        Parameters
        ----------
        path : string
            Path to a target filename.
        key : string
            Datapath within the files.
        storage_options: dict, optional
            Parameters passed to the backend filesystem class.
            Defaults to storage_options set using DatasetFactory.set_default_storage().
        kwargs: dict, optional

        Returns
        -------
        str
            The filename of the HDF5 file created.

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds.to_hdf(path="my/path.h5", key="df")
        """
        if storage_options is None:
            import ads.dataset.factory as factory

            storage_options = factory.default_storage_options
            logger.info("Using default storage options")

        with pd.HDFStore(
            "memory",
            mode="w",
            driver="H5FD_CORE",
            driver_core_backing_store=0,
        ) as hdf_store:
            hdf_store.put(key, self.df, format=kwargs.get("hdf5_format", "fixed"))
            data = hdf_store._handle.get_file_image()

        new_path = (
            path.replace("*", "0")
            if path[-3:] == ".h5"
            else path.replace("*", "0") + ".h5"
        )

        with fsspec.open(
            urlpath=new_path, mode="wb", storage_options=storage_options, **kwargs
        ) as fo:
            fo.write(data)

        return new_path

    @runtime_dependency(module="fastavro", install_from=OptionalDependency.DATA)
    def to_avro(self, path, schema=None, storage_options=None, **kwargs):
        """
        Save data to Avro files.
        Avro is a remote procedure call and data serialization framework developed within Apache's Hadoop project. It
        uses JSON for defining data types and protocols, and serializes data in a compact binary format.

        Parameters
        ----------
        path : string
            Path to a target filename.  May contain a ``*`` to denote many filenames.
        schema : dict
            Avro schema dictionary, see below.
        storage_options: dict, optional
            Parameters passed to the backend filesystem class.
            Defaults to storage_options set using DatasetFactory.set_default_storage().
        kwargs: dict, optional
            See https://fastavro.readthedocs.io/en/latest/writer.html

        Notes
        -----
        Avro schema is a complex dictionary describing the data,
        see https://avro.apache.org/docs/1.8.2/gettingstartedpython.html#Defining+a+schema
        and https://fastavro.readthedocs.io/en/latest/writer.html.
        Its structure is as follows::

            {'name': 'Test',
            'namespace': 'Test',
            'doc': 'Descriptive text',
            'type': 'record',
            'fields': [
                {'name': 'a', 'type': 'int'},
            ]}

        where the "name" field is required, but "namespace" and "doc" are optional
        descriptors; "type" must always be "record". The list of fields should
        have an entry for every key of the input records, and the types are
        like the primitive, complex or logical types of the Avro spec
        (https://avro.apache.org/docs/1.8.2/spec.html).

        Examples
        --------
        >>> ds = DatasetFactory.open("data.avro")
        >>> ds.to_avro("my/path.avro")
        """
        # Get the row by row formatting
        data_row_by_row = []
        for i, row in self.df.iterrows():
            data_row_by_row.append(row.to_dict())
        # Try to auto-generate schema
        if schema is None:
            avro_types = self._convert_dtypes_to_avro_types()
            schema = {"name": self.name, "doc": self.description, "type": "record"}
            fields = []
            ## Add vars
            for col, dtype in avro_types:
                fields.append({"name": col, "type": ["null", dtype]})
            schema["fields"] = fields

        parsed_schema = fastavro.parse_schema(schema=schema)
        new_path = (
            path.replace("*", "0")
            if path[-5:] == ".avro"
            else path.replace("*", "0") + ".avro"
        )
        with fsspec.open(
            new_path, "wb", storage_options=storage_options, **kwargs
        ) as fo:
            fastavro.writer(fo, parsed_schema, data_row_by_row)
        return new_path

    def _convert_dtypes_to_avro_types(self):
        avro_types = []
        for name, dtype in zip(self.dtypes.index, self.dtypes.values):
            if dtype == np.int64:
                avro_dtype = "long"
            elif "int" in str(dtype):
                avro_dtype = "int"
            elif dtype == np.float64:
                avro_dtype = "double"
            elif "float" in str(dtype):
                avro_dtype = "float"
            elif dtype == np.bool_:
                avro_dtype = "boolean"
            else:
                avro_dtype = "string"
            avro_types.append((name, avro_dtype))
        return avro_types

    def astype(self, types):
        """
        Convert data type of features.

        Parameters
        ----------
        types: dict
            key is the existing feature name
            value is the data type to which the values of the feature should be converted.
            Valid data types: All numpy datatypes (Example: np.float64, np.int64, ...)
            or one of categorical, continuous, ordinal or datetime.

        Returns
        -------
        updated_dataset: `ADSDataset`
            an ADSDataset with new data types

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> ds_reformatted = ds.astype({"target": "categorical"})
        """
        return self.__getattr__("astype")(helper.map_types(types))

    def merge(self, data, **kwargs):
        """
        Merges this dataset with another ADSDataset or pandas dataframe.

        Parameters
        ----------
        data : Union[ADSDataset, pandas.DataFrame]
            Data to merge.
        kwargs : dict, optional
            additional keyword arguments that would be passed to underlying dataframe's merge API.

        Examples
        --------
        >>> ds1 = DatasetFactory.open("data1.csv")
        >>> ds2 = DatasetFactory.open("data2.csv")
        >>> ds_12 = ds1.merge(ds2)
        """
        assert isinstance(data, pd.DataFrame) or isinstance(
            data, ADSDataset
        ), "Can only merge datasets if they are of the types pandas or ads"
        df = self.df.merge(data.df if isinstance(data, ADSDataset) else data, **kwargs)
        return self._build_new_dataset(df, progress=utils.get_progress_bar(3))

    """
    Internal methods
    """

    def __getattr__(self, item):
        attr = getattr(self.df, item)
        if callable(attr):
            return self._apply(attr)
        else:
            return attr

    def __getitem__(self, key):
        if isinstance(key, str) or isinstance(key, (tuple, str)):
            return self.df[key]
        else:
            return self._build_new_dataset(self.df[key])

    def _apply(self, func):
        def df_func(*args, _new_target=None, **kwargs):
            has_dataframe_arg = False
            args = list(args)
            for i, arg in enumerate(args):
                if isinstance(arg, ADSDataset) or isinstance(arg, pd.DataFrame):
                    has_dataframe_arg = True
                    # convert any argument that is of type ADSDataset to dataframe. This is useful in delegate calls
                    # like dataset1.concat(dataset2)
                    args[i] = arg.df if isinstance(arg, ADSDataset) else arg

            result = func(*args, **kwargs)

            # return the response as such  if the the result is not a dataframe and it is a read function such as head
            if (
                isinstance(result, pd.DataFrame)
                and func.__name__ not in self.df_read_functions
            ):
                target_name = None
                target_sample_val = None
                if not utils.is_same_class(self, ADSDataset):
                    target_name = (
                        self.target.name if _new_target is None else _new_target
                    )
                    target_sample_val = (
                        self.sampled_df[self.target.name].dropna().values[0]
                    )

                df = result
                n = len(df)
                trans_df = None
                transformed = False
                transformers = []

                # The sampled dataframe needs to be re-generated when this operation involves another dataframe.
                # Also, this kind of transformations cannot be reproduced at the time of scoring
                if not has_dataframe_arg:
                    ft = DataFrameTransformer(
                        func_name=func.__name__,
                        target_name=target_name,
                        target_sample_val=target_sample_val,
                        args=args,
                        kw_args=kwargs,
                    ).fit(result)
                    # transformed is set to false if the method fails to run on pandas dataframe. In this case a new
                    # sampled dataframe is added
                    trans_df, transformed = ft._transform(self.sampled_df.copy())
                    # if the dataset length changes as a result of transformation, these operations need not be added to
                    # pipeline as they do not need to be reproduced at the time of scoring.
                    transformers = (func.__name__, ft) if n == self.shape[0] else []

                init_kwargs = self.init_kwargs.copy()
                if func.__name__ == "astype":
                    if "types" in init_kwargs:
                        init_kwargs["types"] = init_kwargs["types"] + args[0]
                    else:
                        init_kwargs["types"] = args[0]

                # if the transforming function is not supported by pandas dataframe, we need to sample the dask
                # dataframe again to get a new representation
                return self._build_new_dataset(
                    df,
                    sampled_df=df,
                    target=target_name,
                    target_type=TypeDiscoveryDriver().discover(
                        target_name, df[target_name]
                    )
                    if target_name is not None and target_name in df
                    else None,
                    sample=not transformed,
                    transformers=transformers,
                    **init_kwargs,
                )
            return result

        return df_func

    def _handle_key_error(self, args):
        raise ValidationError("Column %s does not exist in data" % str(args))

    def _build_new_dataset(
        self,
        df,
        sampled_df=None,
        target=None,
        target_type=None,
        transformers=[],
        sample=False,
        progress=DummyProgressBar(),
        n=None,
        **init_kwargs,
    ):

        prev_doc_mode = utils.is_documentation_mode()

        set_documentation_mode(False)

        init_kwargs = (
            self.init_kwargs
            if init_kwargs is None or len(init_kwargs) == 0
            else init_kwargs.copy()
        )
        n = len(df) if n is None else n

        # re-calculate sample df if not provided
        if sampled_df is None or sample:
            if progress:
                progress.update("Sampling data")
            sampled_df = generate_sample(
                df,
                n,
                DatasetDefaults.sampling_confidence_level,
                DatasetDefaults.sampling_confidence_interval,
                **init_kwargs,
            )
        else:
            if progress:
                progress.update()
        shape = (n, len(df.columns))
        if not utils.is_same_class(self, ADSDataset) and target is None:
            target = self.target.name

        set_documentation_mode(prev_doc_mode)

        # return a  ADSDataset object if the target has been removed from the dataframe
        if target in sampled_df.columns:
            if progress:
                progress.update("Building new dataset")
            target_type = self.target.type if target_type is None else target_type
            from ads.dataset.factory import DatasetFactory

            new_ds = DatasetFactory._get_dataset(
                df,
                sampled_df,
                target,
                target_type,
                shape,
                progress=progress,
                **init_kwargs,
            )

            new_ds.transformer_pipeline = self._update_transformer_pipeline(
                transformers
            )
            return new_ds
        else:
            if target is not None and not isinstance(progress, DummyProgressBar):
                logger.info(
                    "The target variable does not exist. Use `set_target()` to specify the target."
                )
            if progress:
                progress.update("Building the dataset with no target.")
            dsp = ADSDataset(
                df,
                sampled_df,
                shape,
                progress=progress,
                interactive=False,
                **init_kwargs,
            )
            dsp.transformer_pipeline = self._update_transformer_pipeline(transformers)
            return dsp

    def _validate_feature(self, feature_names):
        if np.isscalar(feature_names):
            feature_names = [feature_names]
        for feature in feature_names:
            if feature not in self.df.columns:
                self._handle_key_error(feature)

    def _update_transformer_pipeline(self, steps=[]):
        if isinstance(steps, tuple):
            steps = [steps]
        if steps is None or len(steps) == 0:
            return copy.deepcopy(self.transformer_pipeline)
        if self.transformer_pipeline is not None:
            transformer_pipeline = TransformerPipeline(
                steps=self.transformer_pipeline.steps + steps
            )
        else:
            transformer_pipeline = TransformerPipeline(steps=steps)
        return transformer_pipeline

    def _get_unique_name(self, name):
        id = (
            uuid.uuid4().hex + "_" + datetime.datetime.utcnow().strftime("%Y%m%d%H%M%S")
        )
        if name == "":
            return id
        return name + "_" + id

    def corr(
        self,
        correlation_methods: Union[list, str] = "pearson",
        frac: float = 1.0,
        sample_size: float = 1.0,
        nan_threshold: float = 0.8,
        overwrite: bool = None,
        force_recompute: bool = False,
    ):
        """
        Compute pairwise correlation of numeric and categorical columns, output a matrix or a list of matrices computed
        using the correlation methods passed in.

        Parameters
        ----------
        correlation_methods: Union[list, str], default to 'pearson'

            - 'pearson': Use Pearson's Correlation between continuous features,
            - 'cramers v': Use Cramer's V correlations between categorical features,
            - 'correlation ratio': Use Correlation Ratio Correlation between categorical and continuous features,
            - 'all': Is equivalent to ['pearson', 'cramers v', 'correlation ratio'].

            Or a list containing any combination of these methods, for example, ['pearson', 'cramers v'].
        frac:
            Is deprecated and replaced by sample_size.
        sample_size: float, defaults to 1.0. Float, Range -> (0, 1]
            What fraction of the data should be used in the calculation?
        nan_threshold: float, default to 0.8, Range -> [0, 1]
            Only compute a correlation when the proportion of the values, in a column, is less than or equal to nan_threshold.
        overwrite:
            Is deprecated and replaced by force_recompute.
        force_recompute: bool, default to be False

            - If False, it calculates the correlation matrix if there is no cached correlation matrix. Otherwise,
              it returns the cached correlation matrix.
            - If True, it calculates the correlation matrix regardless whether there is cached result or not.

        Returns
        -------
        correlation: Union[list, pandas.DataFrame]
            The pairwise correlations as a matrix (DataFrame) or list of matrices
        """
        frac = deprecate_default_value(
            frac,
            None,
            1,
            "<code>frac=None</code> is superseded by <code>sample_size=1.0</code>.",
            FutureWarning,
        )

        if frac != 1.0:
            deprecate_frac = deprecate_variable(
                frac,
                sample_size,
                "<code>frac</code> is superseded by <code>sample_size</code>.",
                DeprecationWarning,
            )
            if sample_size == 1.0:
                sample_size = deprecate_frac

        force_recompute = deprecate_variable(
            overwrite,
            force_recompute,
            f"<code>overwrite=None</code> is deprecated. Use <code>force_recompute</code> instead.",
            DeprecationWarning,
        )
        if sample_size > 1 or sample_size <= 0:
            logger.error("`sample_size` must to be in the range of (0, 1].")
            return
        if nan_threshold > 1 or nan_threshold < 0:
            logger.error("`nan_threshold` must be between 0 and 1 (exclusive).")
            return
        return self._compute_correlation(
            frac=sample_size,
            threshold=nan_threshold,
            force_recompute=force_recompute,
            correlation_methods=correlation_methods,
        )

    def _compute_correlation(
        self,
        frac=1.0,
        threshold=0.8,
        include_n_features=16,
        correlation_methods="pearson",
        force_recompute=False,
    ):
        """
        returns a list of correlation matrix/matrices
        """

        # validate the correlation methods
        correlation_methods = _validate_correlation_methods(correlation_methods)

        # if users choose to sample a frac of the data
        corr_df = self.df if not frac else self.df.sample(frac=frac)

        # return columns by type and filter by threshold
        threshold = threshold * 100
        feature_types_df = pd.DataFrame.from_dict(self.feature_types).T

        # reduce the dim of wide data
        n_rows, n_columns = self.shape

        is_wide_dataset = n_columns >= N_Features_Wide_Dataset

        if is_wide_dataset and include_n_features:
            corr_df, feature_types_df = self._reduce_dim_for_wide_dataset(
                corr_df, feature_types_df, include_n_features
            )

        categorical_columns, continuous_columns, _ = _get_columns_by_type(
            feature_types_df, threshold=threshold
        )

        # get the correlation
        correlation_list = []
        for method in correlation_methods:
            correlation_list.append(
                self._return_correlation(
                    corr_df,
                    method,
                    categorical_columns,
                    continuous_columns,
                    force_recompute,
                )
            )
        return correlation_list[0] if len(correlation_list) == 1 else correlation_list

    def _calc_pearson(self, df: pd.DataFrame, continuous_columns: list) -> pd.DataFrame:
        self._pearson = (
            df[continuous_columns].corr()
            if len(continuous_columns) > 1
            else pd.DataFrame()
        )
        return self._pearson

    def _calc_cramers_v(
        self, df: pd.DataFrame, categorical_columns: list
    ) -> pd.DataFrame:
        self._cramers_v = _cat_vs_cat(df, categorical_columns)
        return self._cramers_v

    def _calc_correlation_ratio(
        self,
        df: pd.core.frame.DataFrame,
        categorical_columns: list,
        continuous_columns: list,
    ) -> pd.DataFrame:
        self._correlation_ratio = _cat_vs_cts(
            df, categorical_columns, continuous_columns
        )
        return self._correlation_ratio

    def _return_correlation(
        self,
        corr_df,
        method,
        categorical_columns,
        continuous_columns,
        force_recompute,
    ):
        if not force_recompute and hasattr(self, "_" + "_".join(method.split())):
            logger.info(
                f"Using cached results for {method} correlation. Use"
                " `force_recompute=True` to override."
            )
            return getattr(self, "_" + "_".join(method.split()))
        else:
            if method == "pearson":
                self._calc_pearson(corr_df, continuous_columns)
                return self._pearson
            elif method == "cramers v":
                self._calc_cramers_v(corr_df, categorical_columns)
                return self._cramers_v
            elif method == "correlation ratio":
                self._calc_correlation_ratio(
                    corr_df, categorical_columns, continuous_columns
                )
                return self._correlation_ratio
            else:
                raise ValueError(f"The {method} method is not supported.")

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def _reduce_dim_for_wide_dataset(
        self, corr_df: pd.DataFrame, feature_types_df: pd.DataFrame, include_n_features
    ):
        min_cores_for_correlation = 2
        n_rows, n_columns = self.shape

        from IPython.core.display import display, HTML

        if utils.get_cpu_count() <= min_cores_for_correlation:
            msg = (
                f"Not attempting to calculate correlations, too few cores ({utils.get_cpu_count()}) "
                f"for wide dataset ({n_columns} columns)"
            )
            display(HTML(f"<li>{msg}</li>"))
            return None, None

        display(HTML(f"<li>detected wide dataset ({n_columns} columns)</li>"))

        if "target" in self.__dict__:
            display(
                HTML(
                    f"<li>feature reduction using mutual information (max {include_n_features} columns)</li>"
                )
            )
            logger.info("Set `include_n_features=None` to include all features.")
            corr_sampled_df = self._find_feature_subset(
                self.sampled_df, self.target.name, include_n_features=include_n_features
            )
            corr_df, feature_types_df = self._update_dataframes(
                corr_sampled_df, corr_df, feature_types_df
            )
        else:
            #
            # in the absense of a target we simply use the first_n
            #
            logger.info(
                f"To include the first {include_n_features} features based on the feature"
                f"importance, use `.set_target`()."
            )
            feature_types_df = feature_types_df[
                (feature_types_df.index.isin(corr_df.columns.values))
                & feature_types_df.type.isin(
                    ["categorical", "ordinal", "continuous", "zipcode"]
                )
            ]
            corr_df = corr_df[feature_types_df.index[:include_n_features]]
            feature_types_df = feature_types_df.iloc[:include_n_features, :]
        return corr_df, feature_types_df

    def _update_dataframes(self, corr_sampled_df, corr_df, feature_types_df):
        """
        update the dataframe and feature types based on the reduced dataframe
        """
        cols = corr_sampled_df.columns.tolist()
        cols.insert(0, cols.pop(cols.index(self.target.name)))
        corr_df_reduced = corr_df[[*cols]]
        feature_types_df_reduced = feature_types_df[feature_types_df.index.isin(cols)]
        return corr_df_reduced, feature_types_df_reduced

    def show_corr(
        self,
        frac: float = 1.0,
        sample_size: float = 1.0,
        nan_threshold: float = 0.8,
        overwrite: bool = None,
        force_recompute: bool = False,
        correlation_target: str = None,
        plot_type: str = "heatmap",
        correlation_threshold: float = -1,
        correlation_methods="pearson",
        **kwargs,
    ):
        """
        Show heatmap or barplot of pairwise correlation of numeric and categorical columns, output three tabs
        which are heatmap or barplot of correlation matrix of numeric columns vs numeric columns using pearson
        correlation method, categorical columns vs categorical columns using Cramer's V method,
        and numeric vs categorical columns, excluding NA/null values and columns which have more than
        80% of NA/null values. By default, only 'pearson' correlation is calculated and shown in the first tab.
        Set correlation_methods='all' to show all correlation charts.

        Parameters
        ----------
        frac: Is superseded by sample_size
        sample_size: float, defaults to 1.0. Float, Range -> (0, 1]
            What fraction of the data should be used in the calculation?
        nan_threshold: float, defaults to 0.8, Range -> [0, 1]
            In the default case, it will only calculate the correlation of the columns which has less than or equal to
            80% of missing values.
        overwrite:
            Is deprecated and replaced by force_recompute.
        force_recompute: bool, default to be False.

            - If False, it calculates the correlation matrix if there is no cached correlation matrix. Otherwise,
              it returns the cached correlation matrix.
            - If True, it calculates the correlation matrix regardless whether there is cached result or not.

        plot_type: str, default to "heatmap"
            It can only be "heatmap" or "bar". Note that if "bar" is chosen, correlation_target also has to be set and
            the bar chart will only show the correlation values of the pairs which have the target in them.
        correlation_target: str, default to Non
            It can be any columns of type continuous, ordinal, categorical or zipcode. When correlation_target is set,
            only pairs that contains correlation_target will show.
        correlation_threshold: float, default to -1
            It can be any number between -1 and 1.
        correlation_methods: Union[list, str], defaults to 'pearson'

            - 'pearson': Use Pearson's Correlation between continuous features,
            - 'cramers v': Use Cramer's V correlations between categorical features,
            - 'correlation ratio': Use Correlation Ratio Correlation between categorical and continuous features,
            - 'all': Is equivalent to ['pearson', 'cramers v', 'correlation ratio'].

            Or a list containing any combination of these methods, for example, ['pearson', 'cramers v'].

        Returns
        -------
        None
        """
        frac = deprecate_default_value(
            frac,
            None,
            1,
            "<code>frac=None</code> is superseded by <code>sample_size=1.0</code>.",
            FutureWarning,
        )
        if frac != 1.0:
            deprecate_frac = deprecate_variable(
                frac,
                sample_size,
                "<code>frac</code> is deprecated. Use <code>sample_size</code> instead.",
                DeprecationWarning,
            )
            if sample_size == 1.0:
                sample_size = deprecate_frac

        feature_types_df = pd.DataFrame.from_dict(self.feature_types).loc["type", :]
        features_list = list(
            feature_types_df[
                feature_types_df.isin(
                    ["categorical", "zipcode", "continuous", "ordinal"]
                )
            ].index
        )
        if plot_type not in ["heatmap", "bar"]:
            raise ValueError('plot_type has to be "heatmap" ' 'or "bar"')

        if plot_type == "bar" and correlation_target is None:
            raise ValueError('correlation_target has to be set when plot_type="bar".')

        if correlation_target:
            if correlation_target not in features_list:
                raise ValueError(
                    "correlation_target has to be in {}.".format(features_list)
                )

        force_recompute = deprecate_variable(
            overwrite,
            force_recompute,
            f"<code>overwrite=None</code> is deprecated. Use <code>force_recompute</code> instead.",
            DeprecationWarning,
        )

        plot_correlation_heatmap(
            ds=self,
            frac=sample_size,
            force_recompute=force_recompute,
            correlation_target=correlation_target,
            plot_type=plot_type,
            correlation_threshold=correlation_threshold,
            nan_threshold=nan_threshold,
            correlation_methods=correlation_methods,
            **kwargs,
        )

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    @runtime_dependency(module="ipywidgets", install_from=OptionalDependency.NOTEBOOK)
    def show_in_notebook(
        self,
        correlation_threshold=-1,
        selected_index=0,
        sample_size=0,
        visualize_features=True,
        correlation_methods="pearson",
        **kwargs,
    ):
        """
        Provide visualization of dataset.

        - Display feature distribution. The data table display will show a maximum of 8 digits,
        - Plot the correlation between the dataset features (as a heatmap) only when all the features are
          continuous or ordinal,
        - Display data head.

        Parameters
        ----------
        correlation_threshold : int, default -1
            The correlation threshold to select, which only show features that have larger or equal
            correlation values than the threshold.
        selected_index: int, str, default 0
            The displayed output is stacked into an accordion widget, use selected_index to force the display to open
            a specific element, use the (zero offset) index or any prefix string of the name (eg, 'corr' for
            correlations)
        sample_size: int, default 0
            The size (in rows) to sample for visualizations
        visualize_features: bool default False
            For the "Features" section control if feature visualizations are shown or not. If not only
            a summary of the numeric statistics is shown. The numeric statistics are also always shows
            for wide (>64 features) datasets
        correlation_methods: Union[list, str], default to 'pearson'

            - 'pearson': Use Pearson's Correlation between continuous features,
            - 'cramers v': Use Cramer's V correlations between categorical features,
            - 'correlation ratio': Use Correlation Ratio Correlation between categorical and continuous features,
            - 'all': Is equivalent to ['pearson', 'cramers v', 'correlation ratio'].

            Or a list containing any combination of these methods, for example, ['pearson', 'cramers v'].
        """

        if not utils.is_notebook():
            print("show_in_notebook called but not in notebook environment")
            return

        n_rows, n_columns = self.shape

        min_sample_size = 10000
        if sample_size == 0:
            sub_samp_size = len(self.sampled_df)
            sub_samp_df = self.sampled_df
        else:
            sub_samp_size = max(min(sample_size, len(self.sampled_df)), min_sample_size)
            sub_samp_df = self.sampled_df.sample(n=sub_samp_size)

        html_summary = ""
        if self.name:
            html_summary += "<h1>Name: %s</h1>" % (self.name)

        # dataset type (problem type)
        html_summary += "<h3>Type: %s</h3>" % self.__class__.__name__

        if self.description:
            html_summary += "<pre>%s</pre>" % self.description
            html_summary += "<hr>"

        html_summary += "<h3>{:,} Rows, {:,} Columns</h3>".format(n_rows, n_columns)
        html_summary += "<h4>Column Types:</h4><UL>"

        for group in Counter(
            [self.feature_types[k].meta_data["type"] for k in self.feature_types]
        ).most_common():
            html_summary += "<LI><b>%s:</b> %d features" % (group[0], group[1])

        html_summary += "</UL>"

        html_summary += """
                <p><b>
                    Note: Visualizations use a sampled subset of the dataset, this is to
                    improve plotting performance. The sample size is calculated to be statistically
                    significant within the confidence level: {} and confidence interval: {}.

                    The sampled data has {:,} rows
                    </b>
                </p>

                <ul>
                    <li>The confidence <i>level</i> refers to the long-term success rate of the
                    method, that is, how often this type of interval will capture the parameter
                    of interest.
                    </li>

                    <li>A specific confidence <i>interval</i> gives a range of plausible values for
                    the parameter of interest
                    </li>
                </ul>

            """.format(
            DatasetDefaults.sampling_confidence_level,
            DatasetDefaults.sampling_confidence_interval,
            sub_samp_df.shape[0],
        )

        html_summary += "</UL>"

        from ipywidgets import widgets

        summary = widgets.HTML(html_summary)

        features = widgets.HTML()
        correlations = widgets.Output()
        warningz = widgets.HTML()

        warningz.value = "Analyzing for warnings..."
        features.value = "Calculating full statistical info..."

        # with correlations:
        #     display(HTML("<li>calculating...</li>"))

        accordion = widgets.Accordion(
            children=[summary, features, correlations, warningz]
        )
        accordion.set_title(0, "Summary")
        accordion.set_title(1, "Features")
        accordion.set_title(2, "Correlations")
        accordion.set_title(3, "Warnings")

        if isinstance(selected_index, str):
            # lookup by title
            possible_titles = [
                accordion.get_title(i) for i in range(len(accordion.children))
            ]
            for i, title in enumerate(possible_titles):
                if title.lower().startswith(selected_index.lower()):
                    selected_index = i
                    break

            if isinstance(selected_index, str):
                # failed to match a title
                logger.info(
                    "`selected_index` should be one of: {}.".format(
                        ", ".join(possible_titles)
                    )
                )
                selected_index = 0

        accordion.selected_index = selected_index

        is_wide_dataset = n_columns >= N_Features_Wide_Dataset

        #
        # set up dataframe to use for correlation calculations
        #

        self.df_stats = self._calculate_dataset_statistics(
            is_wide_dataset, [features, warningz]
        )

        with correlations:
            feature_types_df = pd.DataFrame.from_dict(self.feature_types).loc["type", :]
            if not is_wide_dataset:
                feature_types_df = feature_types_df[
                    self.df_stats["missing"] < len(self.df)
                ]

            frac = kwargs.pop("frac", 1.0)
            overwrite = kwargs.pop("overwrite", None)
            force_recompute = kwargs.pop("force_recompute", False)
            force_recompute = deprecate_variable(
                overwrite,
                force_recompute,
                f"<code>overwrite=None</code> is deprecated. Use <code>force_recompute</code> instead.",
                DeprecationWarning,
            )
            plot_type = kwargs.pop("plot_type", "heatmap")
            correlation_target = kwargs.pop("correlation_target", None)
            nan_threshold = kwargs.pop("nan_threshold", 0.8)
            self.show_corr(
                correlation_threshold=correlation_threshold,
                sample_size=frac,
                force_recompute=force_recompute,
                plot_type=plot_type,
                correlation_target=correlation_target,
                nan_threshold=nan_threshold,
                correlation_methods=correlation_methods,
                **kwargs,
            )

        from IPython.core.display import display

        display(accordion)

        # generate html for feature_distribution & warnings

        accordion.set_title(
            1, f"Features ({n_columns})"
        )  # adjust for datasets with target

        #
        # compute missing value statistics
        # not done for wide datasets
        #

        features.value = self._generate_features_html(
            is_wide_dataset,
            n_columns,
            self.df_stats,
            visualizations_follow=bool(visualize_features),
        )

        warningz.value = self._generate_warnings_html(
            is_wide_dataset, n_rows, n_columns, self.df_stats, warningz, accordion
        )

        if visualize_features and not is_wide_dataset:
            self._visualize_feature_distribution(features)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import os
import subprocess
from contextlib import contextmanager


DEFAULT_KRB5_CONFIG_PATH = "~/.bds_config/krb5.conf"
KRB5_CONFIG = "KRB5_CONFIG"


class KRB5KinitError(Exception):
    """KRB5KinitError class when kinit -kt command failed to generate cached ticket with the keytab file and the krb5 config file."""

    pass


def has_kerberos_ticket():
    """Whether kerberos cache ticket exists."""
    return True if subprocess.call(["klist", "-s"]) == 0 else False


def init_ccache_with_keytab(principal: str, keytab_file: str) -> None:
    """Initialize credential cache using keytab file.

    Parameters
    ----------
    principal: str
        The unique identity to which Kerberos can assign tickets.
    keytab_path: str
        Path to your keytab file.

    Returns
    -------
    None
        Nothing.
    """
    cmd = "kinit -kt %(keytab_file)s  %(principal)s"
    args = {}

    args["principal"] = principal
    args["keytab_file"] = keytab_file

    kinit_proc = subprocess.Popen((cmd % args).split(), stderr=subprocess.PIPE)
    stdout_data, stderr_data = kinit_proc.communicate()

    if kinit_proc.returncode > 0:
        raise KRB5KinitError(stderr_data)


@contextmanager
def krbcontext(
    principal: str, keytab_path: str, kerb5_path: str = DEFAULT_KRB5_CONFIG_PATH
) -> None:
    """A context manager for Kerberos-related actions.
    It provides a Kerberos context that you can put code inside.
    It will initialize credential cache automatically with keytab if no cached ticket exists.
    Otherwise, does nothing.

    Parameters
    ----------
    principal: str
        The unique identity to which Kerberos can assign tickets.
    keytab_path: str
        Path to your keytab file.
    kerb5_path: (str, optional).
        Path to your krb5 config file.

    Returns
    -------
    None
        Nothing.

    Examples
    --------
    >>> from ads.bds.auth import krbcontext
    >>> from pyhive import hive
    >>> with krbcontext(principal = "your_principal", keytab_path = "your_keytab_path"):
    >>>    hive_cursor = hive.connect(host="your_hive_host",
    ...                    port="your_hive_port",
    ...                    auth='KERBEROS',
    ...                    kerberos_service_name="hive").cursor()
    """
    refresh_ticket(principal=principal, keytab_path=keytab_path, kerb5_path=kerb5_path)
    yield


def refresh_ticket(
    principal: str, keytab_path: str, kerb5_path: str = DEFAULT_KRB5_CONFIG_PATH
) -> None:
    """generate new cached ticket based on the principal and keytab file path.

    Parameters
    ----------
    principal: str
        The unique identity to which Kerberos can assign tickets.
    keytab_path: str
        Path to your keytab file.
    kerb5_path: (str, optional).
        Path to your krb5 config file.

    Returns
    -------
    None
        Nothing.

    Examples
    --------
    >>> from ads.bds.auth import refresh_ticket
    >>> from pyhive import hive
    >>> refresh_ticket(principal = "your_principal", keytab_path = "your_keytab_path")
    >>> hive_cursor = hive.connect(host="your_hive_host",
    ...                    port="your_hive_port",
    ...                    auth='KERBEROS',
    ...                    kerberos_service_name="hive").cursor()
    """
    keytab_path = os.path.abspath(os.path.expanduser(keytab_path))
    os.environ[KRB5_CONFIG] = os.path.abspath(os.path.expanduser(kerb5_path))
    if not os.path.exists(os.environ[KRB5_CONFIG]):
        raise FileNotFoundError(f"krb5 config file not found in {kerb5_path}.")
    if not has_kerberos_ticket():
        init_ccache_with_keytab(principal, keytab_path)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import base64
import json
import os
from typing import Dict
from typing import Tuple

import inflection
import yaml
import glob

from ads.common.auth import create_signer
from ads.opctl import logger
from ads.opctl.config.base import ConfigProcessor
from ads.opctl.config.utils import NotSupportedError, convert_notebook
from ads.opctl.constants import (
    ML_JOB_GPU_IMAGE,
    ML_JOB_IMAGE,
)
from ads.opctl.utils import (
    list_ads_operators,
    parse_conda_uri,
    get_region_key,
    get_namespace,
)


YAML_STRUCTURES = {
    "v0": {
        "infrastructure": ["infrastructure"],
        "image": ["execution", "image"],
    },
    "v1": {
        "infrastructure": ["spec", "Infrastructure", "spec"],
    },
}


class ConfigVersioner(ConfigProcessor):
    """
    ads opctl supports multiple yaml file types, kinds and versions.
    Each has its own slightly different yaml structure.
    This goal of this class is to provide a translation from the generic term for a variable
        to it's specific path in it's yaml.
    """

    def __init__(self, config: Dict = None) -> None:
        super().__init__(config)
        self.ads_operators = list_ads_operators()

    def process(self):
        # this function should be run after merging configs
        # conda pack scenarios --
        # - user runs their own scripts and commands -> source_folder + entrypoint + (command)
        # - user runs ADS operator -> name/YAML

        # docker image scenarios --
        # - user runs their own docker image -> image (to run) + (entrypoint) + (command)
        # - user runs ADS operator -> name/YAML

        # TODO: build this out
        logger.debug(f"Config: {self.config}")

        return self



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import, division

import pandas as pd

from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import AbstractTypeDiscoveryDetector
from ads.type_discovery.typed_feature import (
    UnknownTypedFeature,
    CategoricalTypedFeature,
)


class UnknownDetector(AbstractTypeDiscoveryDetector):
    def discover(self, name, series):
        candidate = series.loc[~series.isnull()].iloc[0]

        if series.dtype == "object":
            #
            # if we got all through all the other detectors and it's a string type of feature then we
            # just call it a high dimensional categorical
            #
            return CategoricalTypedFeature.build(name, series)
        else:
            logger.debug(
                "type discovery on column [{}]/[{}] result is Unknown".format(
                    name, series.dtype
                )
            )
            return UnknownTypedFeature.build(name, series)


if __name__ == "__main__":
    dd = UnknownDetector()
    print(dd.discover("unknown", pd.Series([None, "94065", "90210", None])))



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ast
import json
import oci
import os
import pathlib
import re
import shutil

from oci.exceptions import ServiceError
from tempfile import NamedTemporaryFile
from zipfile import ZipFile

from oci.secrets import SecretsClient
from oci.config import from_file

from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common import utils
from ads.common import auth as authutil
from ads.common import oci_client as oc
from ads.vault.vault import Vault


class Connector:
    def __init__(
        self,
        secret_id: str = None,
        key: str = None,
        repository_path: str = None,
        **kwargs,
    ):

        """
        Validate that a connection could be made for the given set of connection parameters, and contruct a Connector object provided that the
        validation is successful.

        Parameters
        ----------
        secret_id: str, optional
            The ocid of the secret to retrieve from Oracle Cloud Infrastructure Vault.
        key: str, optional
            The key to find the database directory.
        repository_path: str, optional
            The local database information store, default to ~/.database unless specified otherwise.
        kwargs: dict, optional
            Name-value pairs that are to be added to the list of connection parameters.
            For example, database_name="mydb", database_type="oracle", username = "root", password = "pwd".

        Returns
        -------
        A Connector object.
        """

        prio_dict = {}

        if kwargs:

            command_creds = {}
            for input_key in kwargs.keys():
                command_creds[input_key] = kwargs.get(input_key)

            # declaring priority order
            prio_dict[1] = command_creds

        # get creds content from vault if secret id is provided
        if secret_id:
            if not bool(re.match("^ocid[0-9]?\.vaultsecret.*", secret_id)):
                raise ValueError(f"{secret_id} is not a valid secret id.")

            auth = authutil.default_signer()
            self.secret_client = oc.OCIClientFactory(**auth).secret

            secret_bundle = self.secret_client.get_secret_bundle(secret_id)
            secret_content = ast.literal_eval(
                Vault._secret_to_dict(secret_bundle.data.secret_bundle_content.content)
            )
            # declaring priority order
            prio_dict[2] = secret_content

        # get creds content from local if key is provided
        repository_path = _get_repository_path(repository_path=repository_path)
        if not os.path.exists(repository_path):
            raise ValueError(f"{repository_path} does not exist.")

        if key:
            if _not_valid_key(key=key):
                raise ValueError(f"{key} is not a valid directory name.")
            db_path = _get_db_path(repository_path=repository_path, key=key)
            if not os.path.exists(db_path):
                raise ValueError(f"{db_path} does not exist.")
            local_content = get_repository(key=key, repository_path=repository_path)
            # declaring priority order
            prio_dict[3] = local_content

        # Combine dictionary with priority using ** operator
        config = {}
        for k in reversed(sorted(prio_dict.keys())):
            config.update(**prio_dict[k])
        self.config = config

        # check database types
        valid_database_types = ["oracle"]

        if "database_type" not in self.config:
            raise ValueError(
                f"The database_type needs to be specified. "
                f"Valid database types are {valid_database_types}"
            )

        if self.config["database_type"] not in valid_database_types:
            raise ValueError(
                f"{self.config['database_type']} is not a valid database type. "
                f"Valid database types are {valid_database_types}"
            )

        if self.config["database_type"] == "oracle":
            connector = OracleConnector(self.config)
            self.uri = connector.uri
            self.engine = connector.engine

    def connect(self):
        return self.engine.connect()

    def __enter__(self):
        self.db_connection = self.engine.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self

    def __repr__(self):  # pragma: no cover
        return str(self.engine.url)


class OracleConnector:
    @runtime_dependency(module="sqlalchemy", install_from=OptionalDependency.DATA)
    def __init__(self, oracle_connection_config):
        self.config = oracle_connection_config

        # sanity check on valid keys before making a connection
        valid_keys = ["username", "password", "database_name"]
        for vk in valid_keys:
            if vk not in self.config.keys():
                raise ValueError(f"{vk} is a required parameter to connect.")

        self.uri = _create_connection_str(config=self.config)
        self.engine = sqlalchemy.create_engine(self.uri)


def update_repository(
    value: dict, key: str, replace: bool = True, repository_path: str = None
) -> dict:
    """
    Saves value into local database store.

    Parameters
    ----------
    value: dict
        The values to store locally.
    key: str
        The key to find the local database directory.
    replace: bool, default to True
        If set to false, updates the stored value.
    repository_path:str: str, optional
        The local database store, default to ~/.database unless specified otherwise.

    Returns
    -------
    A dictionary of all values in the repository for the given key.
    """
    if _not_valid_key(key=key):
        raise ValueError(f"{key} is not a valid directory name.")

    # make local database directory
    repository_path = _get_repository_path(repository_path=repository_path)
    pathlib.Path(repository_path).mkdir(parents=True, exist_ok=True)

    db_path = _get_db_path(repository_path=repository_path, key=key)
    pathlib.Path(db_path).mkdir(parents=True, exist_ok=True)

    db_config_path = os.path.join(db_path, "config.json")

    if not replace:
        value = _update(value, db_config_path)

    with open(db_config_path, "w") as fp:
        json.dump(value, fp)

    return value


def _update(new_value, db_config_path):
    # update existing key's values if found different and add new key-value pairs
    with open(db_config_path) as f:
        old_value = json.load(f)
    prio_dict = {1: old_value, 2: new_value}
    return {**prio_dict[2], **prio_dict[1]}


def get_repository(key: str, repository_path: str = None) -> dict:
    """
    Get all values from local database store.

    Parameters
    ----------
    key: str
        The key to find the database directory.
    repository_path: str, optional
        The path to local database store, default to ~/.database unless specified otherwise.

    Returns
    -------
    A dictionary of all values in the store.
    """
    if _not_valid_key(key=key):
        raise ValueError(f"{key} is not a valid directory name.")

    # check whether repository_path exists
    repository_path = _get_repository_path(repository_path=repository_path)
    if not os.path.exists(repository_path):
        raise ValueError(f"{repository_path} does not exist.")
    # check whether db_path exists
    db_path = _get_db_path(repository_path=repository_path, key=key)
    if not os.path.exists(db_path):
        raise ValueError(f"{db_path} does not exist.")

    db_config_path = os.path.join(db_path, "config.json")
    with open(db_config_path) as f:
        return json.load(f)


def import_wallet(wallet_path: str, key: str, repository_path: str = None) -> None:
    """
    Saves wallet to local database store.
    Unzip the wallet zip file, update sqlnet.ora and store wallet files.

    Parameters
    ----------
    wallet_path: str
        The local path to the downloaded wallet zip file.
    key: str
        The key to find the database directory.
    repository_path: str, optional
        The local database store, default to ~/.database unless specified otherwise.

    """
    if _not_valid_key(key=key):
        raise ValueError(f"{key} is not a valid directory name.")

    # checking paths are valid
    repository_path = _get_repository_path(repository_path=repository_path)
    db_path = _get_db_path(repository_path=repository_path, key=key)

    if not os.path.exists(db_path):
        raise ValueError(f"{db_path} does not exist.")
    if not os.path.exists(wallet_path):
        raise ValueError(f"{wallet_path} does not exist.")

    # Create a ZipFile Object and load wallet zip in it
    with ZipFile(wallet_path, "r") as zipObj:
        # Extract all the contents of zip file in current directory
        zipObj.extractall(
            os.path.join(os.path.expanduser("~"), f"{repository_path}/{key}")
        )

    # Add TNS_ADMIN to the environment
    os.environ["TNS_ADMIN"] = db_path

    # Update the sqlnet.ora
    _update_sqlnet(db_path=db_path)

    # Update the config.json file so that the 'database_type' is set to oracle.
    db_config_path = os.path.join(db_path, "config.json")
    value = _update({"database_type": "oracle"}, db_config_path)
    with open(db_config_path, "w") as fp:
        json.dump(value, fp)


def _update_sqlnet(db_path):
    sqlnet_path = os.path.join(db_path, "sqlnet.ora")
    sqlnet_original_path = os.path.join(db_path, "sqlnet.ora.original")
    sqlnet_backup_path = os.path.join(db_path, "sqlnet.ora.backup")

    if not os.path.exists(sqlnet_original_path):
        shutil.copy(sqlnet_path, sqlnet_original_path)
    if os.path.exists(sqlnet_path):
        shutil.copy(sqlnet_path, sqlnet_backup_path)
    sqlnet_re = re.compile(
        '(WALLET_LOCATION\s*=.*METHOD_DATA\s*=.*DIRECTORY\s*=\s*")(.*)(".*)',
        re.IGNORECASE,
    )
    tmp = NamedTemporaryFile()

    with open(sqlnet_path, "rt") as sqlnet:
        for line in sqlnet:
            tmp.write(
                bytearray(
                    sqlnet_re.subn(r"\1{}\3".format(db_path), line)[0], encoding="utf-8"
                )
            )
    tmp.flush()
    shutil.copy(tmp.name, sqlnet_path)
    tmp.close()


def _get_repository_path(repository_path):
    return (
        os.path.join(os.path.expanduser("~"), ".database")
        if repository_path is None
        else repository_path
    )


def _get_db_path(repository_path, key):
    return os.path.join(repository_path, key)


def _not_valid_key(key):
    return bool(re.search("[../.]", key))


def _create_connection_str(config):
    return (
        "oracle+cx_oracle://"
        + config["username"]
        + ":"
        + config["password"]
        + "@"
        + config["database_name"]
    )
